import requests
import numpy as np
import json
from dateutil.tz import *
from http import HTTPStatus
import urllib.parse

DATETIME_FORMATS = {
    'local': '%Y-%m-%dT%H:%M:%S%z',
}
LOCAL_TIME_ZONE = tzlocal()

INTERVAL = {
    '15min': 900,
    'hour': 3600,
    'day': -1,
    'week': -2,
    'month': -3    
}

HOSTNAME = 'https://localhost:9099'

class BaiException(Exception):
    """Raise bai exception"""

def resolve_holiday(country: str, region:str, dates:list[dict], token:str):
    payload = {
        'location': {
            'country': country,
            'region': region,
        },
        'dates': dates
    }

    response = requests.post(
        '{}/service/intelli/v1/resolve/holiday'.format(HOSTNAME), 
        data=json.dumps(payload), 
        headers={        
            'Content-Type': 'application/json',    
            'x-access-token': token
    }, verify=False)

    response.raise_for_status()

    body = json.loads(response.text)

    if body['error'] != None:
        raise Exception(response.text['error'])
    
    return body['data']


def serialize_interpolation(interpolation:dict) -> str:
    interpolateMethod = ''

    if 'method' in interpolation:    
        # Test for existence due to data schema update
        diff = interpolation['method']['diff'] if 'diff' in interpolation['method'] else ''
        statistic = interpolation['method']['statistic'] if 'statistic' in interpolation['method'] else ''    

        interpolateMethod = "{}{}{}{}{}".format(diff, interpolation['method']['name'], interpolation['method']['type'], interpolation['method']['fill'], statistic)            
    return interpolateMethod

def fetch_endpoint(pmModel:dict, token: str, as_localhost: bool = True) -> str:
    url = "{}/service/local-process-manager/v1/instances/servers?filterProcessName=DMS".format(HOSTNAME)
    if pmModel['processManagerUuid'] != 'local-process-manager':
        url = "{}/service/{}/service/local-process-manager/v1/instances/servers?filterProcessName=DMS".format(HOSTNAME, pmModel['processManagerUuid'])

    response = requests.get(url, headers={
        'Content-Type': 'application/json',
        'x-access-token': token,
    }, verify=False)
    response.raise_for_status()

    responseBody = json.loads(response.content)

    server = None
    for instance in responseBody['instances']:
        if instance['uuid'] == pmModel['instanceUuid']:
            for process in instance['processes']:
                if pmModel['processUuid'] != '':
                    if process['uuid'] == pmModel['processUuid']:
                        server = process['servers'][0]
                else:
                    if process['name'] == pmModel['processName']:
                        server = process['servers'][0]                    

    if server == None:
        raise Exception("Server or process uuid not found - Instance might be inactive or you might have copied app/BAI data from elsewhere")
    
    scheme = "http"
    if server['location']['scheme'] == "wss":
        scheme = "https"

    host = server['location']['hostIp']
    if as_localhost:
        host = 'localhost'

    endpoint = "{}://{}:{}{}".format(scheme, host, server['location']['port'], server['location']['basePath'])
    return endpoint

def fetch_hist_data(pmModel:dict, interpolation:dict, interval:int, token: str, start_date: str, end_date: str = None) -> list:

    histData = {
        "start": start_date,
        "interval": interval,
    }

    histData['interpolateTemplate'] = interpolation['template'] if 'template' in interpolation else ''
    histData['interpolateMethod'] = serialize_interpolation(interpolation)
    histData['filterParams'] = interpolation['filterParams'] if 'filterParams' in interpolation else {}
    histData['normalizeParams'] = interpolation['normalizeParams'] if 'normalizeParams' in interpolation else []
    histData['prospective'] = interpolation['prospective'] if 'prospective' in interpolation else False

    """
    Remove interpolateMethod if empty and template not empty since
    an empty string on interpolateMethod defaults to prevNextLinearFill
    and will override the interpolation method defined in interpolateTemplate
    """
    if histData['interpolateTemplate'] != '' and histData['interpolateMethod'] == '':
        del histData['interpolateMethod']
    
    if end_date:
        histData['end'] = end_date

    endpoint = fetch_endpoint(pmModel, token)
    payload = {
        'url': endpoint,
        'payload': {
            "get": [{
                "path": pmModel['path'],
                "showExtInfos": ["name", "state", "unit"],
                "histData": histData
            }] 
        }
    }

    url = '{}/service/dms-proxy/v1/json_data'.format(HOSTNAME)
    if pmModel['processManagerUuid'] != 'local-process-manager':
        url = '{}/service/{}/service/dms-proxy/v1/json_data'.format(HOSTNAME, pmModel['processManagerUuid'])

    response = requests.post(url, data=json.dumps(payload), headers={
    'Content-Type': 'application/json',
    'x-access-token': token
    },
    verify=False)

    response.raise_for_status()
    data = json.loads(response.text)

    if data['payload']['get'][0]['code'] != 'ok':
        raise Exception(data)

    return data['payload']['get'][0]['histData']
    
def get_closest_index(value, bin_range) -> int:
    """
    Given a sorted range, seek the position of {value} and return the index
    """    
    for i, n in enumerate(bin_range):
        if value >= n and value <= bin_range[i + 1]:
            return i    
    
    
def get_bin_range(base: int, step: int, min_val:int, max_val:int) -> list:
    """
    Generate a sorted range between the {min} {max} values starting at {base} with steps {step}
    """
    try:
        forwards = np.arange(base, max_val+(step), step)
        backwards = np.arange(base, min_val-(step), -step)
    except:
        raise_bai_exception('Binning has failed verify values (base: {} max_val: {} min_val: {} step: {})'.format(base, max_val, min_val, step))        

    # Combine backwards with forward values [-4, -2, 0, 2, 4]
    bin_range = np.concatenate((np.flip(backwards), forwards)).tolist()

    # Ticket 45108
    # min: 0, max:0, base: 0 results in range [0.0] instead of [0.0, 0.0]        
    if len(bin_range) > 2:
        bin_range = list(dict.fromkeys(bin_range))
    elif len(bin_range) < 2:
        raise_bai_exception('bin_range: {} must contain at least two values'.format(bin_range))

    return bin_range

def fetch_log_service(token:str, process_manager:str, query:dict = None) -> list:

    url = '{}/service/logservice/api/v1/logServices'.format(process_manager)
    if query != None:
        query_string = urllib.parse.urlencode(query)
        url = '{}?{}'.format(url, query_string)

    response = requests.get(url, headers={        
        'x-access-token': token
    },
    verify=False)
    response.raise_for_status()

    body = json.loads(response.text)
    return body['data']

def raise_bai_exception(msg: str, parameters: dict = {}):
    raise BaiException('{{ "errorMsg": "{}", "parameters": {} }}'.format(msg, json.dumps(parameters)))

def fetch_config_model(uuid:str, token:str, base_url:str) -> dict:
    uri = "/service/rest/v1/filedb/app/BAI/configModel/{}.json".format(uuid)
    return request_apigateway(uri, base_url, headers={'x-access-token': token})

def fetch_widget(uuid:str, token:str, base_url:str) -> dict:
    uri = "/service/rest/v1/filedb/app/BAI/widgets/{}.yaml".format(uuid)
    return request_apigateway(uri, base_url, headers={'x-access-token': token})

def fetch_outputs(uuid:str, token:str, base_url:str) -> dict:
    uri = "/service/rest/v1/filedb/app/BAI/outputs/{}.json".format(uuid)
    return request_apigateway(uri, base_url, headers={'x-access-token': token})

def request_apigateway(uri:str, base_url:str, headers:dict = {}, method:str = 'get'):
    response = getattr(requests, method)("{}{}".format(base_url, uri), headers=headers, verify=False)
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        raise Exception('Invalid token')        
    """
    If not json file, return response content directly.
    Specially for files/data requests
    """
    if 'application/json' in response.headers['Content-Type']:
        file_db = json.loads(response.content)
        if ('error' in file_db):
            if response.status_code == HTTPStatus.NOT_FOUND:
                raise FileNotFoundError(uri)    
            else:
                raise Exception(file_db['error'])
        data = file_db['data']['data']
    else:
        data = response.content
    return data