
import os
import sys
from datetime import datetime, timedelta
import argparse
import pandas as pd
import json
import traceback
import numpy as np

# Load parent dir for intelli_utilities module
# Append maxpermissible values folder
currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)

# Append scripts folder
parentdir = os.path.dirname(parentdir)
sys.path.append(parentdir)

from intelli_utilities import fetch_widget, fetch_config_model, DATETIME_FORMATS, fetch_endpoint, fetch_log_service, serialize_interpolation, resolve_holiday, HOSTNAME
from utils import get_model_params
import requests
from lib.Thresholds import Thresholds
from DataProcessor import DataProcessor
from maxpermissiblevalues.metrics.metric_utils import get_node_unique_identifier

import logging
logging.basicConfig(filename='log/cron.log', encoding='utf-8', level=logging.ERROR)

BI_SYSTEM_PATHS = {
    'expected': 'BISystem:{}:Expected',
    'high': {
        'limit': 'BISystem:{}:LimitHigh',
        'violation': 'BISystem:{}:LimitHigh:violation',
        'alarm': 'BISystem:{}:LimitHigh:Alarm'
    },
    'low': {
        'limit': 'BISystem:{}:LimitLow',
        'alarm': 'BISystem:{}:LimitLow:Alarm',
        'violation': 'BISystem:{}:LimitLow:violation'
    }
}

def set_trend_histdata(path:str, datetime:str, value:any) -> list:
    #Hist data attribute will create TRD points
    return {
            'path': path, 
            'type': 'double',
            'create': True,
            'value': value,
            'histData': [{
                datetime: value
            }],
    }


def get_log_template(path:str, value:str, type:str, stamp:str):
    return {
        'sourceId': path,
        'field': type,
        'oldValue': '',
        'newValue': value,
        'private': False,
        'logType': 'intelli',
        'createdAt': stamp # "2024-09-04T14:55:37+02:00"
    }


def get_alarm_template(path:str, create:bool, stamp:str, value:any)  -> list:
    return [
        {'path': path, 'create': create,  'value': value, 'type': 'bool', 'stamp': stamp},
        {'path': '{}:ALM:Alarm1'.format(path), 'create': create, 'type': None},
        {'path': '{}:ALM:Alarm1:Activ'.format(path), 'create': create, 'value': True, 'type': 'bool'},
        {'path': '{}:ALM:Alarm1:ExtAlmText'.format(path), 'create': create, 'value': '', 'type': 'string'},
        {'path': '{}:ALM:Alarm1:Format'.format(path), 'create': create, 'type': 'string', 'value': ''},
        # {'path': '{}:ALM:Alarm1:Format'.format(path), 'create': create, 'value': 'Alarm1@Anlage / #c / #N / #V^NAME / #VComment / #Z(kommt:geht:quit) #u', 'type': 'string'},
        {'path': '{}:ALM:Alarm1:Group'.format(path), 'create': create, 'value': 0, 'type': 'int'},
        {'path': '{}:ALM:Alarm1:Pri'.format(path), 'create': create, 'value': 5, 'type': 'int'},
        {'path': '{}:ALM:Alarm1:Prot'.format(path), 'create': create, 'value': 'ALARM', 'type': 'string'},
        {'path': '{}:ALM:Alarm1:Text'.format(path), 'create': create, 'value': '', 'type': 'string'},
        {'path': '{}:ALM:Alarm1:Type'.format(path), 'create': create, 'value': '>=', 'type': 'string'},
        {'path': '{}:ALM:Alarm1:Value'.format(path), 'create': create, 'value': 1.000, 'type': 'double'},
    ]

def set_holidays(df:pd.DataFrame, compare_holidays:list) -> pd.DataFrame:
    holiday_days = [datetime.strptime(date, DATETIME_FORMATS['local']).date() for date in compare_holidays]
    df['holiday'] = 0
    for row in df.itertuples():
        if row[0].date() in holiday_days:
            df.at[row.Index, 'holiday'] = 1
    return df

def nodes_to_dataframe(nodes:list, config_model:dict, dates:list, dp:DataProcessor, token:str) -> pd.DataFrame:
    data_df = pd.DataFrame()
    for node in nodes:
        col_name = get_node_unique_identifier(node['pmModel'])
        hist_data = dp.get_hist_data_from_dates(node['pmModel'], node['interpolation'], config_model['interval'], dates)
        df = dp.convert_histdata_to_dataframe(hist_data, col_name)

        if 'holidayLocation' in config_model:
            holiday_location = config_model['holidayLocation']
            if holiday_location['country']:
                holiday_dates = resolve_holiday(holiday_location['country'], holiday_location['region'], dates, token)
                df = set_holidays(df, holiday_dates)

        if 'bins' in node:
            if 'base' in node['bins'] and 'step' in node['bins']:
                df = dp.binarize_data(df, col_name, node['bins'])
        
        data_df = data_df.join(df, how='outer', sort=True)

    data_df = dp.add_temporal_variables(data_df)
    return data_df

def persist_config_model(config_model:dict, args):
    # persist next training date
    uri = '{}/service/rest/v1/filedb/app/BAI/configModel/{}.json'.format(args.process_manager, config_model['uuid'])
    requests.post(uri,
        json=config_model, 
        headers={
            'x-access-token': args.token
        },
        verify=False            
    )

def get_training_dates(config_model:dict, now:datetime):
    training_dates = []
    match config_model['trainMethod']:
        case 'movingWindows':
            start_date = now - timedelta(days=config_model['periodsDay']['train'])
            training_dates = [{
                'start': start_date.isoformat(),
                'end': now.isoformat()
            }]                            
        case 'isolatedWindows':
            training_dates = config_model['trainingDates']
        case _:
            raise Exception('trainMethod does not exist')
    return training_dates

def get_transform_df(config_model:dict, now, token:str):
    """
    Fetches the current datapoint value
    For counters, the use the last value and the latest one the hour value to perform differentiation
    """
    start_date = None
    match config_model['interval']:
        case 3600:
            start_date = now - timedelta(hours=1)
        case -1:
            start_date = now - timedelta(days=1)
        case _:
            raise Exception('interval does not exist')

    start_date_time = start_date.isoformat()
    end_date_time = now.isoformat()

    data_df = pd.DataFrame()
    current_value = None

    for node in config_model['nodes']:
        url = '{}/service/dms-proxy/v1/json_data'.format(HOSTNAME)
        if node['pmModel']['processManagerUuid'] != 'local-process-manager':
            url = '{}/service/{}/service/dms-proxy/v1/json_data'.format(HOSTNAME, node['pmModel']['processManagerUuid'])
                    
        col_name = get_node_unique_identifier(node['pmModel'])
        endpoint = fetch_endpoint(node['pmModel'], token)

        body = {
            'url': endpoint,
            'payload': {
                "get": [{
                        "path": node['pmModel']['path'],
                        "showExtInfos": ["name", "unit"],
                        "histData": {
                            'interval': config_model['interval'],
                            'start': start_date_time,
                            'end': end_date_time,
                        }
                    },                     
            ]}
        }
        
        if 'template' in node['interpolation'] and node['interpolation']['template'] not in ['', 'custom']:                
            body['payload']['get'][0]['histData']['interpolateTemplate'] = node['interpolation']['template'] 
        else:
            body['payload']['get'][0]['histData']['interpolateMethod'] = serialize_interpolation(node['interpolation']['method'])
        
        logging.error('body')
        logging.error(json.dumps(body))

        response = requests.post(url, data=json.dumps(body), headers={
            'Content-Type': 'application/json',
            'x-access-token': token
        },
        verify=False)

        response.raise_for_status()
        data = json.loads(response.text)
        logging.error('response body')
        logging.error(response.text)

        get = data['payload']['get']

        for i, hist in enumerate(get[0]['histData']):
            logging.error('values {} {}: {}'.format(i, list(hist.values())[0], list(hist.keys())[0]))
            # print('On the hour {} {}: {}'.format(i, list(hist.values())[0], list(hist.keys())[0]))
        
        logging.error('final values')
        logging.error('start_date_time: {}, end_date_time: {}'.format(start_date_time, end_date_time))
        current_value = list(get[0]['histData'][-1].values())[0]
        logging.error('current values: {}'.format(json.dumps(get[0]['histData'])))
        logging.error('current value rounded 3 dec: {}'.format(np.round(current_value, 3)))


        df = pd.DataFrame({
            col_name: np.round(current_value, 3)
        }, index=[now])

        if 'bins' in node:
            if 'base' in node['bins'] and 'step' in node['bins']:
                df = dp.binarize_data(df, col_name, node['bins'])
        data_df = data_df.join(df, how='outer', sort=True)

    data_df = dp.add_temporal_variables(data_df)

    dates = [{
        'start': start_date_time,
        'end': end_date_time
     }]
    
    if 'holidayLocation' in config_model:
        holiday_location = config_model['holidayLocation']
        if holiday_location['country']:
            holiday_days = resolve_holiday(holiday_location['country'], holiday_location['region'], dates, token)
            data_df = set_holidays(data_df, holiday_days)


    return data_df

if __name__ == '__main__':
    # Set cli args
    parser = argparse.ArgumentParser()
    parser.add_argument('--widget-uuid')
    parser.add_argument('--action-uuid')
    parser.add_argument('--token', help='Security token for services')
    parser.add_argument('--process-manager', help='Base url of the process manager')
    parser.add_argument('--now')
    args = parser.parse_args()


    logging.error('----- START DEBUGGER -------')
    logging.error('executed at {}'.format(datetime.now().strftime(DATETIME_FORMATS['local'])))
    logging.error('widget uuid {}'.format(args.widget_uuid))
    logging.error('action uuid {}'.format(args.action_uuid))

    try:
        now_str = args.now
        now = datetime.strptime(args.now, DATETIME_FORMATS['local'])
        dp = DataProcessor(args.token, args.process_manager)

        whois = 'BI'
        user = 'bi_cron'

        try:
            widget = fetch_widget(args.widget_uuid, args.token, args.process_manager)
            action = None
            for item in widget['actions']:
                if item['uuid'] == args.action_uuid:
                    action = item
            config_model = fetch_config_model(widget['configModelUuid'], args.token, args.process_manager)
        except requests.exceptions.RequestException as e:
            raise SystemExit(e)
        
        # pick dependent var pmModel
        dependent_pm_model = None
        for node in config_model['nodes']:
            unique_name = get_node_unique_identifier(node['pmModel'])
            if unique_name == config_model['dependentVariables'][0]:
                dependent_pm_model = node['pmModel']
                break    

        thresholds = Thresholds()
        thresholds.parse_config_model(config_model)
        # Todo set model parameters! model_params = get_model_params(config_model)    
        threshold_instance = thresholds.create_thresold_instance(config_model['task']['algorithm']['method'], config_model['dependentVariables'], config_model['independentVariables'], config_model['task']['algorithm']['model'])    
        try:
            threshold_instance.load_model(action['modelUuid'], {'x-access-token': args.token}, args.process_manager)                            
        except FileNotFoundError:
            logging.error("model being trained")
            # Save model if does not exist yet
            training_dates = get_training_dates(config_model, now)
            if config_model['trainMethod'] == 'movingWindows':
                next_training_date = now + timedelta(days=config_model['periodsDay']['retrain'])
                config_model['nextTraining'] = next_training_date.strftime(DATETIME_FORMATS['local'])            
                persist_config_model(config_model, args)

            train_df = nodes_to_dataframe(config_model['nodes'], config_model, training_dates, dp, args.token)
            # Train and persist model
            threshold_instance.fit(train_df)            
            threshold_instance.save_model(action['modelUuid'], {'x-access-token': args.token}, args.process_manager)

        if config_model['trainMethod'] == 'movingWindows':
            # Verify is model retraining is needed
            next_training_date = datetime.strptime(config_model['nextTraining'], DATETIME_FORMATS['local']) 
            if now > next_training_date:
                training_dates = get_training_dates(config_model, now)

                next_training_date = now + timedelta(days=config_model['periodsDay']['retrain'])
                config_model['nextTraining'] = next_training_date.strftime(DATETIME_FORMATS['local'])            
                persist_config_model(config_model, args)

                train_df = nodes_to_dataframe(config_model['nodes'], config_model, training_dates, dp, args.token)
                # Train and persist model
                threshold_instance.fit(train_df)
                threshold_instance.save_model(action['modelUuid'], {'x-access-token': args.token}, args.process_manager)            


        data_df = get_transform_df(config_model, now, args.token)    
        logging.error('transform_data: {}'.format(data_df.to_json()))
        result = threshold_instance.transform(data_df)

        # Read events from change logs
        upper_event_count = config_model['eventCount']['upper']
        lower_event_count = config_model['eventCount']['lower']

        delta_param_upper = {}
        delta_param_lower = {}
        match config_model['interval']:
            case 3600:
                delta_param_upper = {'hours': upper_event_count}
                delta_param_lower = {'hours': lower_event_count}
            case -1:
                delta_param_upper = {'days': upper_event_count}
                delta_param_lower = {'days': lower_event_count}
            case _:
                raise Exception('interval not found')

        start_date = (now - timedelta(**delta_param_upper)).strftime(DATETIME_FORMATS['local'])
        start_date = start_date[:-2] + ':' + start_date[-2:]

        event_upper = False
        alarm_upper = False
        upper_logs = []
        if config_model['trespassing']['upper'] == True:
            if result['dist_upper'].values[0] > 0:
                log_filter = {
                    'logType': 'intelli',
                    'sourceId': '{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['high']['violation'].format(config_model['pathId'])),
                    'createdAtFrom': start_date,
                }
                upper_logs = fetch_log_service(args.token, args.process_manager, log_filter)
                if len(upper_logs) >= upper_event_count:
                    alarm_upper = True
                else:
                    event_upper = True

        event_lower = False
        alarm_lower = False
        lower_logs = []
        if config_model['trespassing']['lower'] == True:
            if result['dist_lower'].values[0] > 0:
                log_filter = {
                    'logType': 'intelli',
                    'sourceId': '{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['low']['violation'].format(config_model['pathId'])),
                    'createdAtFrom': start_date,
                }
                lower_logs = fetch_log_service(args.token, args.process_manager, log_filter)
                if len(lower_logs) >= lower_event_count:
                    alarm_lower = True
                else:
                    event_lower = True

        dependent_node_col_name = get_node_unique_identifier(dependent_pm_model)
        current_value = str(result[dependent_node_col_name].values[0])

        logging.error("change_logs_upper: {}".format(json.dumps(upper_logs)))
        logging.error("change_logs_lower: {}".format(json.dumps(lower_logs)))

        logging.error("dist_upper: {}".format(result['dist_upper'].values[0]))
        logging.error("dist_lower: {}".format(result['dist_lower'].values[0]))

        logging.error("Y_pred: {}".format(result['Y_pred'].values[0]))
        logging.error("th_upper: {}".format(result['upper'].values[0]))
        logging.error("th_lower: {}".format(result['lower'].values[0]))
        logging.error("current_value: {}".format(current_value))

        logging.error("event_upper: {}".format(event_upper))
        logging.error("event_lower: {}".format(event_lower))
        logging.error("alarm_lower: {}".format(alarm_lower))
        logging.error("alarm_upper: {}".format(alarm_upper))


        # Hanlde preAlarm and alarms for plotting purpose with help of the type
        change_log = None
        if event_upper == True or alarm_upper == True:
            upper_type = 'preAlarm' if event_upper == True else 'alarm'
            change_log = get_log_template('{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['high']['violation'].format(config_model['pathId'])), current_value, upper_type, now_str) 

        if event_lower == True or alarm_lower == True:
            lower_type = 'preAlarm' if event_lower == True else 'alarm'
            change_log = get_log_template('{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['low']['violation'].format(config_model['pathId'])), current_value, lower_type, now_str)

        # Handle only alarms. Has no effect on alarm plotting
        set_alarm = []
        if alarm_lower or alarm_upper:
            if alarm_lower:
                set_alarm = set_alarm + get_alarm_template('{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['low']['alarm'].format(config_model['pathId'])), True, now_str, alarm_lower)
            if alarm_upper:
                set_alarm = set_alarm + get_alarm_template('{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['high']['alarm'].format(config_model['pathId'])), True, now_str, alarm_upper)
        
        if alarm_upper == False and event_upper == False:
           set_alarm.append({'path': '{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['high']['alarm'].format(config_model['pathId'])), 'type': 'bool', 'value': False})
        if alarm_lower == False and event_lower == False:
           set_alarm.append({'path': '{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['low']['alarm'].format(config_model['pathId'])), 'type': 'bool', 'value': False})           

        dms_proxy_service_url = '{}/service/dms-proxy/v1/json_data'.format(HOSTNAME)
        if dependent_pm_model['processManagerUuid'] != 'local-process-manager':
            dms_proxy_service_url = '{}/service/{}/service/dms-proxy/v1/json_data'.format(HOSTNAME, dependent_pm_model['processManagerUuid'])
        endpoint = fetch_endpoint(dependent_pm_model, args.token)

        dms_set = [
            set_trend_histdata('{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['expected'].format(config_model['pathId'])), now_str, result['Y_pred'].values[0]),
            set_trend_histdata('{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['high']['limit'].format(config_model['pathId'])), now_str, result['upper'].values[0]),
            set_trend_histdata('{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['low']['limit'].format(config_model['pathId'])), now_str, result['lower'].values[0]),                                
        ]

        dms_set = dms_set + set_alarm
        response = [{
            'url': dms_proxy_service_url,
            'data': {
                'url': endpoint,
                'payload': {
                    'whois': whois,
                    'user': user,
                    'set': dms_set
                },      
            }
        }]
        
        if change_log != None:
            response.append({
                'url': '{}/service/logservice/api/v1/logServices'.format(HOSTNAME),
                'data': change_log
            })

        logging.error(json.dumps(response))
        print(json.dumps(response))
    except Exception as e:
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]        
        error = "{} {} {} {} {}".format(e, exc_type, fname, exc_tb.tb_lineno, traceback.format_exc())
        logging.error(error)