import traceback
import sys, os
import json
import pandas as pd
import numpy as np
import argparse
import requests
import warnings
warnings.filterwarnings("ignore")

# Append maxpermissible values folder
currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)

from intelli_utilities import LOCAL_TIME_ZONE

from sklearn.linear_model import *
from sklearn.tree import DecisionTreeRegressor, ExtraTreeRegressor
from sklearn.neighbors import KNeighborsRegressor
from sklearn.ensemble import AdaBoostRegressor, BaggingRegressor, ExtraTreesRegressor, GradientBoostingRegressor, HistGradientBoostingRegressor, RandomForestRegressor
from sklearn.gaussian_process.kernels import Matern
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.kernel_ridge import KernelRidge
from sklearn.svm import LinearSVR, SVR, NuSVR
from sklearn.neural_network import MLPRegressor
from sklearn.dummy import DummyRegressor
from mlinsights.mlmodel import PiecewiseRegressor
from metrics.metric_utils import print_stream, get_colname_list, get_node_unique_identifier
from metrics.calculate_metrics import calculate_metrics
from sklearn.model_selection import GridSearchCV
import xgboost as xgb
from sklearn.preprocessing import KBinsDiscretizer
import jsonpickle
from lib.RegressionThresholds import *
from lib.Thresholds import Thresholds
from datetime import datetime
import logging
logging.basicConfig(filename='log/hyperparamter_tuning.log', encoding='utf-8', level=logging.ERROR)

hypertuning_params = {
    'AdaBoostRegressor': {
        'model': AdaBoostRegressor,
        'search_grid': {
            'n_estimators': [20, 50, 100, 500, 1000],
            'learning_rate':[.001, 0.01, .1],
            'loss': ['linear', 'square', 'exponential'],
            'random_state':[1]
        }
    },
    'BaggingRegressor': {
        'model': BaggingRegressor,        
        'search_grid': {
            'n_estimators': [20, 50, 100, 200, 400, 500, 700, 1000],
            'max_samples': [x / 10 for x in range(1, 11)],
            'max_features': [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.90, 1.0],
            'bootstrap': [True, False],
            'bootstrap_features': [True, False],
            'oob_score': [True, False],
        }            
    },
    'DecisionTreeRegressor':{
        'model': DecisionTreeRegressor,
        'search_grid': {
            'criterion':['squared_error', 'friedman_mse', 'absolute_error', 'poisson'],              
            'splitter': ['best', 'random'],
            'max_depth': [None, 10, 20, 30], 
            'min_samples_split': [2, 5, 10],
            'min_samples_leaf': [1, 2, 4],
            'max_features': ['auto', 'sqrt', 'log2']               
        }
    },
    'ExtraTreesRegressor':{
        'model': ExtraTreesRegressor,        
        'search_grid': {
            'criterion':['squared_error', 'friedman_mse', 'absolute_error', 'poisson'],              
            'n_estimators': [20, 50, 100, 200, 400, 500, 700],
            'max_depth': [None, 10, 20, 30], 
            'min_samples_split': [2, 5, 10], 
            'min_samples_leaf': [1, 2, 4]           
        }
    },
    'GammaRegressor': { # Where is Gamma Regressor in result list??
        'model': GammaRegressor,
        'search_grid': {
            'alpha':[0, 1, 3, 5, 10],
            'solver': ['lbfgs', 'newton-cholesky'], 
            'max_iter': [20, 50, 100, 150, 200], 
            'tol': [1e-8, 1e-5, 1e-4, 1e-2],
        }            
    },
    'GaussianProcessRegressor': {
        'model': GaussianProcessRegressor,        
        'search_grid': {
            'kernel': [None, Matern()],
            'alpha': [1e-10, 1e-8, 1e-5],
            'n_restarts_optimizer': [0, 1, 5],
            'normalize_y': [False, True],   
            'copy_X_train': [False, True]
        }            
    }, 
    'GradientBoostingRegressor': {
        'model': GradientBoostingRegressor,        
        'search_grid': {
            'loss': ['squared_error', 'huber'],
            'learning_rate': [0.001, 0.01, 0.1, 0.2],             
            'n_estimators': [50, 100, 200, 500],
            'max_depth': [3, 4, 5],
            'min_samples_split': [2, 5, 10],
            'min_samples_leaf': [1, 2, 4]                      
        }
    },
    'HistGradientBoostingRegressor': {
        'model': HistGradientBoostingRegressor,        
        'search_grid': {
            'loss': ['squared_error', 'quantile'],
            'learning_rate': [0.001, 0.01, 0.1, 0.2],
            'max_iter': [50, 100, 200],
            'max_leaf_nodes': [15, 27, 45],
            'max_depth': [None, 5, 10],
            'min_samples_leaf': [2, 4],
            'l2_regularization': [0.0, 0.1, 0.2]
        }
    },
    'KNeighborsRegressor': {
        'model': KNeighborsRegressor,        
        'search_grid': {
            'n_neighbors': [3, 5, 7, 9],
            'weights': ['uniform', 'distance'],
            'p': [1, 2]
        }
    },
    'KernelRidge': {
        'model': KernelRidge,        
        'search_grid': {
            'alpha': [0.1, 1.0, 10.0],
            'kernel': ['linear', 'rbf', 'poly'],
            'gamma': [0.01, 0.1, 1.0],
            'degree': [2, 3, 4],
        }
    },
    'Lars': {
        'model': Lars,        
        'search_grid': {            
            'fit_intercept': [True, False],
            'n_nonzero_coefs': [np.inf, 100, 500, 800],
            'eps': [1e-4, 1e-3, 1e-2],
            'copy_X': [True, False],            
        }
    },
    'LarsCV': {
        'model': LarsCV,        
        'search_grid': {
            'max_n_alphas': [300, 600, 1000, 1200],
            'fit_intercept': [True, False],
            'max_iter': [1, 20, 50, 100, 150, 200],
            'eps': [1e-4, 1e-3, 1e-2],
            'copy_X': [True, False],
        }            
    },
    'Lasso': {
        'model': Lasso,        
        'search_grid': {
            'alpha': [5, 10, 50, 100],
            'fit_intercept': [True, False],            
            'tol': [1e-4, 1e-3, 1e-2],
            'max_iter': [300, 500, 1000, 1500],
            'copy_X': [True, False],
        }            
    },     
    'LassoCV': {
        'model': LassoCV,        
        'search_grid': {
            'n_alphas': [50, 100, 200, 300],
            'fit_intercept': [True, False],            
            'tol': [1e-4, 1e-3, 1e-2],
            'max_iter': [300, 500, 1000, 1500],
            'copy_X': [True, False],
        }            
    },    
    'LassoLars': {
        'model': LassoLars,        
        'search_grid': {
            'alpha': [0.5, 1.0, 2.0],
            'fit_intercept': [True, False],            
            'eps': [1e-4, 1e-3, 1e-2],
            'max_iter': [100, 300, 500, 1000],
            'copy_X': [True, False],
            'positive': [True, False]            
        }            
    },    
    'LassoLarsCV': {
        'model': LassoLarsCV,        
        'search_grid': {
            'fit_intercept': [True, False],            
            'eps': [1e-4, 1e-3, 1e-2],
            'max_iter': [100, 300, 500, 1000],
            'copy_X': [True, False],
            'positive': [True, False]            
        }            
    },
    'LassoLarsIC': {
        'model': LassoLarsIC,        
        'search_grid': {
            'criterion': ['aic', 'bic'],
            'fit_intercept': [True, False],
            'eps': [1e-4, 1e-3, 1e-2],            
            'max_iter': [100, 300, 500, 1000],
            'copy_X': [True, False],
            'positive': [True, False]
        }
    },
    'LinearRegression': {
        'model': LinearRegression,
        'search_grid': {
            'fit_intercept': [True, False],
            'copy_X': [True, False],
            'positive': [True, False]
        }
    },
    'LinearSVR': {
        'model': LinearSVR,        
        'search_grid': {
            'C': [0.0, 0.1, 1.0, 10.0],
            'epsilon': [0, 0.1, 0.2, 0.5],
            'tol': [1e-4, 1e-3, 1e-2],
            'fit_intercept': [True, False],            
            'loss': ['epsilon_insensitive', 'squared_epsilon_insensitive'],
            'dual': ['auto']
        }
    },
    'MLPRegressor': {
        'model': MLPRegressor,        
        'search_grid': {
            'hidden_layer_sizes': [(50,), (100,), (50, 50), (100, 50, 25)],
            'activation': ['relu', 'tanh'],
            'solver': ['adam', 'sgd'],
            'alpha': [0.0001, 0.001, 0.01],
            'learning_rate': ['constant', 'invscaling', 'adaptive']
        }
    },
    'NuSVR': {
        'model': NuSVR,        
        'search_grid': {
            'nu': [0.1, 0.3, 0.5, 0.8],
            'C': [1, 10, 100],
            'degree': [2, 3, 5],
            'gamma': ['scale', 'auto'],
            'kernel': ['linear', 'rbf']
        }
    },
    'OrthogonalMatchingPursuit': { # Working partially
        'model': OrthogonalMatchingPursuit,        
        'search_grid': {
            'n_nonzero_coefs': [None, 1, 2],
            'fit_intercept': [True, False],
        }
    },
    'OrthogonalMatchingPursuitCV': {
        'model': OrthogonalMatchingPursuitCV,        
        'search_grid': {
            'n_nonzero_coefs': [5, 10, 15],
            'normalize': [True, False],
            'cv': [3, 5, 10],                    
        }
    },
    'RANSACRegressor': {
        'model': RANSACRegressor,        
        'search_grid': {
            'min_samples': [None, 0.1, 0.3, 0.5, 1.0],
            'max_trials': [50, 100, 200],
            'residual_threshold': [5, 10, 20]
        }
    },
    'RidgeCV': {
        'model': RidgeCV,        
        'search_grid': {
            'alphas': [
                [0.1, 1.0, 10.0, 100.0],
                [0.2, 2.0, 20.0, 200.0],
            ],
            'gcv_mode': ['auto', 'svd', 'eigen'],
            'fit_intercept': [True, False],                
        }
    },
    'BayesianRidge': {
        'model': BayesianRidge,        
        'search_grid': {
            'max_iter': [100, 200, 300, 500],
            'tol': [1e-7, 1e-5, 1e-3, 1e-2],            
            'alpha_1': [1e-8, 1e-6, 1e-4, 1e-2],
            'alpha_2': [1e-8, 1e-6, 1e-4, 1e-2],
            'lambda_1': [1e-8, 1e-6, 1e-4, 1e-2],
            'lambda_2': [1e-8, 1e-6, 1e-4, 1e-2],
            'fit_intercept': [False, True],
        }
    },   
    'SVR': {
        'model': SVR,        
        'search_grid': {
            'C': [0.1, 1, 10, 100],
            'kernel': ['rbf', 'sigmoid'],
            'gamma': ['scale', 'auto', 0.1, 1, 10],
            'degree': [2, 3, 4]      
        }
    },
    'TheilSenRegressor': {
        'model': TheilSenRegressor,        
        'search_grid': {
            'fit_intercept': [True, False],
            'max_subpopulation': [1e2, 1e3, 1e4], 
            'max_iter': [100, 300, 500],
            'n_subsamples': [None, 5, 10, 40],
            'tol': [1e-5, 1e-3, 1e-2],                          
        }
    },
    'PiecewiseRegressor': {
        'model': PiecewiseRegressor,        
        'search_grid': {
            'binner': [
                KBinsDiscretizer(n_bins=4),                
                KBinsDiscretizer(n_bins=5),
                KBinsDiscretizer(n_bins=10),
                KBinsDiscretizer(n_bins=15),
                KBinsDiscretizer(n_bins=20),
                KBinsDiscretizer(n_bins=30),
                DecisionTreeRegressor(max_depth=None, min_samples_leaf=10, min_samples_split=2),
                DecisionTreeRegressor(max_depth=5, min_samples_leaf=1, min_samples_split=2),
                DecisionTreeRegressor(max_depth=10, min_samples_leaf=2, min_samples_split=5),
                DecisionTreeRegressor(max_depth=15, min_samples_leaf=4, min_samples_split=10),
                DecisionTreeRegressor(max_depth=None, min_samples_leaf=1, min_samples_split=10),
                DecisionTreeRegressor(min_samples_leaf=300),
            ]
        }
    },
    'XGBoost': {
        'model': xgb.XGBRegressor,        
        'search_grid': {
            'learning_rate': [0.01, 0.1, 0.2],
            'n_estimators': [50, 100, 200, 1000],
            'max_depth': [3, 5, 7],
            'subsample': [0.8, 1.0],
            'colsample_bytree': [0.8, 1.0],
            'reg_alpha': [0, 0.1, 0.5],
            'reg_lambda': [1.0, 1.5, 2.0]
        }            
    },
    'RandomForestRegressor': {
        'model': RandomForestRegressor,        
        'search_grid': {
            'bootstrap': [True, False],
            'max_depth': [10, 20, 30, 40, 50, 60, 70, 80, 90, 100, None],
            'max_features': ['auto', 'sqrt'],
            'min_samples_leaf': [1, 2, 4],
            'min_samples_split': [2, 5, 10],
            'n_estimators': [200, 400, 600, 800, 1000, 1200, 1400, 1600, 1800, 2000]}           
    },
    'Ridge': {
        'model': Ridge,        
        'search_grid': {
            'alpha': [1, 5, 10],
            'copy_X': [False, True],
            'fit_intercept': [False, True],
            'tol': [0.001, 0.01, 0.1, 0.2],             
            'solver': ['auto', 'svd']            
        }
    },
    'ElasticNetCV': {
        'model': ElasticNetCV,        
        'search_grid': {
            'l1_ratio': [0.2, 0.5, 0.7, 0.9],
            'eps': [1e-4, 1e-3, 1e-2],
            'n_alphas': [50, 100, 200, 300],            
            'fit_intercept': [False, True],
            'max_iter': [100, 500, 1000, 1200, 1500],
            'tol': [1e-8, 1e-4, 1e-2],
            'selection': ['cyclic', 'random'],
        }
    },
    'ARDRegression': {
        'model': ARDRegression,        
        'search_grid': {
            'max_iter': [100, 200, 300, 500],
            'tol': [1e-7, 1e-5, 1e-3, 1e-2],            
            'alpha_1': [1e-8, 1e-6, 1e-4, 1e-2],
            'alpha_2': [1e-8, 1e-6, 1e-4, 1e-2],
            'lambda_1': [1e-8, 1e-6, 1e-4, 1e-2],
            'lambda_2': [1e-8, 1e-6, 1e-4, 1e-2],
            'fit_intercept': [False, True],
        }
    },
    'HuberRegressor': {
        'model': HuberRegressor,        
        'search_grid': {
            'epsilon': [1, 1.35, 1.85, 2, 5],
            'max_iter': [50, 100, 300, 500],
            'alpha': [0.0001, 0.001, 0.01, 0.1],
            'fit_intercept': [False, True],
            'tol': [1e-7, 1e-5, 1e-3, 1e-2],                
        }
    },
    'DummyRegressor': {
        'model': DummyRegressor,        
        'search_grid': {
            'strategy': ['mean', 'median', 'constant'],
        }
    },
    'ExtraTreeRegressor': {
        'model': ExtraTreeRegressor,        
        'search_grid': {
            'criterion': ['squared', 'friedman_mse'],
            'splitter': ['random', 'best'],
            'max_depth': [None, 10, 20, 30],
            'min_samples_split': [2, 5, 10],
            'min_samples_leaf': [1, 2, 4]              
        }
    },      
}

if __name__ == "__main__":
    response = {
        'data': '',
        'error': '',
    }
                 
    try:                
   
        parser = argparse.ArgumentParser()
        parser.add_argument('--jobUuid', help='Uuid of the page to load')
        parser.add_argument('--context', help='execution target. ie. local, cluster')        
        parser.add_argument('--token', help='Security token for services')
        parser.add_argument('--processManager', help='Base url of the process manager')        
        args = parser.parse_args()
        
        JOB_UUID = args.jobUuid
        EXECUTION_CONTEXT = args.context                
        
        # Read input data
        SPOOL_DIR_UUID = f"/opt/intelli-service/spool/{JOB_UUID}"        
        if EXECUTION_CONTEXT == 'local':
            SPOOL_DIR_UUID =  f"spool/{JOB_UUID}"
        
        input_file = open(f"{SPOOL_DIR_UUID}/input.json")
        input_data = json.load(input_file)
        input_file.close()
        
        job = input_data['job']
        job_config = job['config']
        main_job_config = job_config['mainJobConf']
        main_result_config = job_config['resultConf']
        result_uuid = job_config['resultUuid']

        data_df = pd.DataFrame(input_data['dataFrame'])
        data_df = data_df.set_index('dateTime', drop=False)
        data_df.index = pd.to_datetime(data_df.index)                

        dataframes = []
        for date in main_job_config['trainingDates']:
            dt_start = datetime.fromisoformat(date['start']).astimezone(LOCAL_TIME_ZONE)
            dt_end = datetime.fromisoformat(date['end']).astimezone(LOCAL_TIME_ZONE)
            dataframes.append(data_df[dt_start:dt_end])
        train_data_df = pd.concat(dataframes)

        test_data_df = pd.DataFrame()
        if len(main_job_config['measureDates']) > 0:
            dataframes = []
            for date in main_job_config['measureDates']:
                dt_start = datetime.fromisoformat(date['start']).astimezone(LOCAL_TIME_ZONE)
                dt_end = datetime.fromisoformat(date['end']).astimezone(LOCAL_TIME_ZONE)
                dataframes.append(data_df[dt_start:dt_end])
            test_data_df = pd.concat(dataframes)

        dependent_node_colname = get_node_unique_identifier(main_result_config['dependentVar']['pmModel'])
        explanatory_var_name_list = get_colname_list(main_result_config['explanatoryVariables'])

        model_name = main_result_config['task']['model']
        grid_search = GridSearchCV(
            hypertuning_params[model_name]['model'](), 
            param_grid=hypertuning_params[model_name]['search_grid'], 
            scoring='neg_mean_squared_error',
            cv=3,
            n_jobs=-1,
        )
        
        # For isolated windows concat train and test to use all data for gridsearch
        concated_data = pd.concat([train_data_df, test_data_df])
    
        grid_search.fit(concated_data[explanatory_var_name_list].values, concated_data[dependent_node_colname].values)
               
        model = hypertuning_params[model_name]['model'](**grid_search.best_params_)
        
        threshold_instance = RegressionThresholds(dependent_variables=[dependent_node_colname], explanatory_vars=explanatory_var_name_list, model=model)
        thresholds = Thresholds()
 
        if main_job_config['trainMethod'] == 'isolatedWindows':                
            results_df = thresholds.train_static_data(threshold_instance, train_data_df, test_data_df)

        if main_job_config['trainMethod'] == 'movingWindows':
            results_df, _, _ = thresholds.train_rolling_data(threshold_instance, train_data_df, main_job_config['periodsDay']['train'], main_job_config['periodsDay']['retrain'], True)
        
        mseEScaled = None
        if not results_df.empty:
            metric = calculate_metrics(results_df, dependent_node_colname)
            # mseEScaled can be very high due to binning error in #45108
            # mseEScaled can be 0.0 due to most Y_pred = nan. Probably a consequent of to many missing infos (minBinElements)
            if np.isnan(metric) == False:            
                if metric > 10000 or metric != 0.0:
                   mseEScaled = metric
                    
        picklable = [
            'GaussianProcessRegressor',
            'PiecewiseRegressor',
            'RANSACRegressor',
        ]
        
        main_result_config['gridSearch'] = {
            'bestParams':  '',
            'bestParamsEncoded': '',
            'bestScore': grid_search.best_score_,
            'tunedMetric': mseEScaled
        }
        
        if model_name in picklable:
            main_result_config['gridSearch']['bestParamsEncoded'] = jsonpickle.encode(grid_search.best_params_)            
        else:
            main_result_config['gridSearch']['bestParams'] = grid_search.best_params_                        
                                                  
        response['data'] = {
            'resultUuid': result_uuid,
            'gridSearch': main_result_config['gridSearch'],            
        }
        print_stream(json.dumps(response))        

        # Delete tuning job
        response = requests.delete("{}/service/rest/v1/filedb/app/BAI/jobs/{}.yaml".format(args.processManager, args.jobUuid), headers={
            'x-access-token': args.token
        }, verify=False)
        response.raise_for_status()        
        
    except Exception as e:
        response = {
            'error': None
        }
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]        
        response['error'] = "{} {} {} {} {}".format(e, exc_type, fname, exc_tb.tb_lineno, traceback.format_exc())
        logging.error(response['error'])
        print_stream(json.dumps(response))