from lib.baseClass import BaseThresholdClass
import pandas as pd
import numpy as np
import requests
import json
import skops.io as sio

class RegressionThresholds(BaseThresholdClass):
    """This class provides threholds based on residuals from a linear regression model.

    Thresholds are calculated by (fit-method)
    1. Fitting training data to a linear model
    2. Calculate the residuals between the model and data
    3. Calculate the quantiles over the residuals
    The calculated thresholds are applied (tranform-method) by:
    1. calculating the expected value through the linear regression model
    2. adding the quantile values to the expected value.
    """
    def __init__(self,
                 model,                 
                 dependent_variables: list,
                 explanatory_vars:list, 
                 q_lower=0.01, 
                 q_upper=0.99,                 
                 **kwargs):

        self._model = model
        self.model_info = {'params': {}, 'info': {}}
        self._q_lower = q_lower
        self._q_upper = q_upper        
        self._quantiles = None
        self._residuals = None
        self.dependent_variable = dependent_variables[0]
        self._explanatory_vars = explanatory_vars.copy()
        self._kwargs = kwargs

    def fit(self, data:pd.DataFrame) -> dict:
        data = data.copy() # Um SettingWithCopy-Fehler zu vermeiden        
        data.dropna(inplace=True)
        
        X = data[self._explanatory_vars].values
        Y = data.loc[:, self.dependent_variable].values
        self._model.fit(X, Y)
        Y_pred = np.round(self._model.predict(X).flatten(), self._decimals)
        self._residuals = data.loc[:, self.dependent_variable] - Y_pred
        self._quantiles = self._residuals.quantile(q=[self._q_lower, self._q_upper])

    def transform(self, data:pd.DataFrame) -> pd.DataFrame:
        if self._model is None:     
            raise Exception('Missing Model')
            
        data = data.copy() # Um SettingWithCopy-Fehler zu vermeiden
        data = data.dropna()

        X = data[self._explanatory_vars].values
        predictions = self._model.predict(X).flatten()
        Y_pred = np.round(predictions, self._decimals)
        
        data['Y_pred'] = Y_pred # predictions

        lower_quantile = self._quantiles.loc[self._q_lower]
        if (isinstance(lower_quantile, pd.Series)):
            lower_quantile = lower_quantile.iloc[0]

        upper_quantile = self._quantiles.loc[self._q_upper]
        if (isinstance(upper_quantile, pd.Series)):
            upper_quantile = upper_quantile.iloc[0]

        data['lower'] = np.round(data['Y_pred'] + lower_quantile, self._decimals)
        # data.loc[(data['lower'] < 0), 'lower'] = 0 # set negative thresholds to 0
        data['upper'] = np.round(data['Y_pred']+ upper_quantile, self._decimals)
        # data.loc[(data['upper'] < 0), 'upper'] = 0 # set negative thresholds to 0
        
        # get minimal distance to threshold
        data = self._get_min_dist(data)
        
        return data
    
    def save_model(self, dir_name:str, headers:dict, process_manager:str):
        """
        :param dir_name: save location. Usually yo want to set config_model_uuid
        :param token
        :param base_url
        """    
        model = sio.dumps(self._model)

        requests.post("{}/service/rest/v1/files/data/app/BAI/models/{}/model.sav".format(process_manager, dir_name),
            files={'file': model}, 
            headers=headers,
            verify=False            
        )

        requests.post("{}/service/rest/v1/filedb/app/BAI/models/{}/quantiles.json".format(process_manager, dir_name),
            json=self._quantiles.to_json(), 
            headers=headers,
            verify=False
        )        
        return model
    
    def load_model(self, dir_name:str, headers:dict, process_manager:str):
        """
        :param dir_name: save location. Usually yo want to set config_model_uuid
        :param token
        :param process_manager
        """           
        data = self.request_apigateway(
            '/service/rest/v1/files/data/app/BAI/models/{}/model.sav'.format(dir_name), 
            process_manager, 
            headers
        )
        # TODO: Test untrusted types list. First test ended up in empty array
        unknown_types = sio.get_untrusted_types(data=data)
        self._model = sio.loads(data, unknown_types)    
        
        quantiles_str = self.request_apigateway(
            '/service/rest/v1/filedb/app/BAI/models/{}/quantiles.json'.format(dir_name),
            process_manager,
            headers
        )

        self._quantiles = pd.Series(json.loads(quantiles_str))        
        self._quantiles.index = self._quantiles.index.astype('float64')