from lib.baseClass import BaseThresholdClass
import pandas as pd
import numpy as np
import requests

class ResidualThresholds(BaseThresholdClass):
    """This class provides threholds based on residuals.

    Thresholds are calculated by (fit-method)
    1. grouping training data into different folds (accoring to 'group_by') and
    2. determining the mean/median per group.
    3. The mean/median per group is subracted from the corresponding values to
       result with residuals.
    4. The quantiles _over all residuals_ are then added to the mean/median per group
       to obtain thresholds.
    The calculated thresholds are applied (tranform-method) to the identically grouped evaluation data.
    """
    def __init__(self, 
                 dependent_variables:list,     
                 explanatory_vars:list,
                 use_stats='median',
                 q_lower=0.01, 
                 q_upper=0.99,
                 min_no_elements_per_bin=3,
                 **kwargs):
        """
        :param group_by: list containing keys to group data
        :param use: Type of statistic to calculate residuals per bin, options ('mean', 'median')
        """                   
        self.dependent_variable = dependent_variables[0]
        self._explanatory_vars = self.select_binarized_data(explanatory_vars.copy())
        self._use_stats = use_stats
        self._q_lower = q_lower
        self._q_upper = q_upper
        self._min_no_elements_per_bin = min_no_elements_per_bin
        self._stats_df= None
        self.model_info = None
        self._residuals = None
        self.missing_data = []
            
    def fit(self, data:pd.DataFrame) -> dict:
                                        
        """Calculate thresholds based on provided data"""
        data = data.copy()    
        
        # check for sufficient data
        if not self._is_sufficient_data_available(data):
            raise ValueError(f"Cannot do training. Provided data for explanatory vars '{self._explanatory_vars}' does not contain sufficient valid data.") 

        grouped = data.groupby(by=self._explanatory_vars)
        
        self._stats_df = grouped[[self.dependent_variable]].agg(self._use_stats)
        # get size of each group
        counts_df = pd.DataFrame(grouped.size(), columns=['size'])
        self._stats_df = self._stats_df.join(counts_df)
        # calculate thresholds from residuals
        residuals = []
        for keys, group in grouped:
            # to calculate residuals, only include groups with more than 
            # 'min_no_elements_per_bin' assuming that the mean or median 
            # will not meaningful/stable otherwise
            if group.shape[0] >= self._min_no_elements_per_bin:
                residuals.append(group[self.dependent_variable] - self._stats_df.loc[keys, self.dependent_variable])
            else:          
                self.missing_data.append({'name': 'tooFewBinElements', 'args': { 'binAmount': group.shape[0], 'binMinimum': self._min_no_elements_per_bin}})

        try:
            self._residuals = pd.concat(residuals, axis=0)
            quantiles = self._residuals.quantile([self._q_lower, self._q_upper])
        except Exception:
            raise ValueError("Algorithmus benötigt mehr Training Daten.")


        lower_quantile = quantiles.loc[self._q_lower]
        if (isinstance(lower_quantile, pd.Series)):
            lower_quantile = lower_quantile.iloc[0]

        upper_quantile = quantiles.loc[self._q_upper]
        if (isinstance(upper_quantile, pd.Series)):
            upper_quantile = upper_quantile.iloc[0]

        self._stats_df['lower'] = np.round(self._stats_df[self.dependent_variable] + lower_quantile, self._decimals)
        self._stats_df.loc[(self._stats_df['lower'] < 0) , 'lower'] = 0
        self._stats_df['upper'] = np.round(self._stats_df[self.dependent_variable] + upper_quantile, self._decimals)
        self._stats_df.loc[(self._stats_df['upper'] < 0) , 'upper'] = 0
        self._stats_dict = self._stats_df.to_dict()
        self.model_info={
            'params': self._stats_df[self.dependent_variable].to_dict(),
            'info'  : self._stats_df['size'].to_dict(),
             }
        return self.model_info

    def _get_range(self, data_row:pd.Series):
        """get acceptable range for the corresponding tuple."""
        idx_tuple = tuple(data_row[self._explanatory_vars].values)

        # handle case where group_by contains only single element
        idx_tuple = idx_tuple[0] if len(self._explanatory_vars) == 1 else idx_tuple
        if idx_tuple in self._stats_df.index:
            # in case the number of elements is smaller than min_no_elements_per_bin
            #TODO: improve error message in case _min_no_elements_per_bin is never reached!
            if self._stats_df.loc[idx_tuple, "size"] <= self._min_no_elements_per_bin:   
                return [np.nan, -np.inf, np.inf, self._stats_dict['size'][idx_tuple]]
            else:
                return [self._stats_dict[self.dependent_variable][idx_tuple],
                        self._stats_dict['lower'][idx_tuple],
                        self._stats_dict['upper'][idx_tuple],
                        self._stats_dict['size'][idx_tuple]
                       ]
        else:
            # in case there was no bin for this combination in the training data
            return [np.nan, -np.inf, np.inf, 0.0]

    def transform(self, data:pd.DataFrame) -> pd.DataFrame:        
        
        data = data.dropna()    
        data = data.copy()        
        
        # check that self._stats_df is populated
        if self._stats_df is None:
            raise ValueError("Call 'fit' before 'transform'")
        
        if len(data) == 0:
            raise ValueError("Cannot do prediction as input contains not a single valid datapoint.")        

        # get thesholds
        data = data.join(data.apply(self._get_range, axis=1, result_type='expand'), rsuffix='_r')
        data = data.rename(columns={0 : 'Y_pred', 1 : 'lower', 2: 'upper', 3 : 'size'})

        # get minimal distance to threshold 
        data = self._get_min_dist(data)
        return data
    
    def save_model(self, dir_name:str, headers:dict, process_manager:str):
        """
        :param dir_name: save location. Usually yo want to set config_model_uuid
        :param token
        :param process_manager
        """   
        model = self._stats_df.reset_index().to_dict('records')
        requests.post(
            "{}/service/rest/v1/filedb/app/BAI/models/{}/thresholds.json".format(process_manager, dir_name),
            json=model, 
            headers=headers,
            verify=False
        )
        return model

    def load_model(self, dir_name:str, headers:dict, process_manager:str):
        """
        :param dir_name: save location. Usually yo want to set config_model_uuid
        :param token
        :param process_manager
        """                   
        data = self.request_apigateway(
            '/service/rest/v1/filedb/app/BAI/models/{}/thresholds.json'.format(dir_name),
            process_manager, 
            headers,
        )        
        self._stats_df = pd.DataFrame(data)
        self._stats_df.set_index(self._explanatory_vars, inplace=True)
        self._stats_dict = self._stats_df.to_dict()
