import pandas as pd
import numpy as np
from dateutil.relativedelta import relativedelta
from dateutil.tz import tzlocal
import traceback

import os, sys

currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)

# Append scripts folder
parentdir = os.path.dirname(parentdir)
sys.path.append(parentdir)

from intelli_utilities import raise_bai_exception
from utils import get_all_regression_estimators, get_model_params

from outlierDetection import find_outliers_extended

from lib.ResidualThresholds import *
from lib.RegressionThresholds import *

from metrics.calculate_metrics import *

METHODS_CLASS_DICT = {
    'thresholdsRegression': RegressionThresholds,        
    'thresholdsGrouping' : ResidualThresholds,    
    # 'QuantileThresholds' : QuantileThresholds,
}    

class Thresholds():
    
    def __init__(self):
        self._regressor_model_list = get_all_regression_estimators()            
        

    def get_model(self, name):
        return self._regressor_model_list[name]

    def train_static_data(self, threshold_instance, training_data:pd.DataFrame, testing_data:pd.DataFrame) -> pd.DataFrame:
        transf_df = pd.DataFrame()
        try:
            threshold_instance.fit(training_data)
            transf_df = threshold_instance.transform(testing_data)
        except ValueError:
            pass
        
        return transf_df
    

    def train_rolling_data(self, threshold_instance, data:pd.DataFrame, train_period:int, retrain_period: int, ignore_outliers:bool):
        results_df = pd.DataFrame()

        # Allows training and retrainig area rendering in chart
        training_dates = []
        retraining_dates = []

        # Parameters & Data
        training_period_dateOffset = pd.DateOffset(days=train_period)
        retraining_period_dateOffset = pd.DateOffset(days=retrain_period)

        training_start_date = data.index.min() + relativedelta(days=+train_period)
        training_end_date = data.index.max() + relativedelta(days=+retrain_period)
        
        # Time index for retraining
        # Set replace(tzinfo=tzlocal()) since we dont have DateTimeIndex anymore
        date_range = pd.date_range(start=training_start_date.replace(tzinfo=tzlocal()), end=training_end_date.replace(tzinfo=tzlocal()), freq=retraining_period_dateOffset)
        # date_range = pd.date_range(start=training_start_date, end=training_end_date, freq=retraining_period_dateOffset)
        df_list = []
        outliers_df_list = []
        errors = ""
        previous_testing_data = []

        # Iteration über die Zeitsprünge von 'RETRAINING_PERIOD' (z.B. 3 Monate) beginnend vom ersten Zeitpunkt+TRAINING_PERIOD
        for i_date in date_range[:-1]:
            training_data = data.loc[i_date-training_period_dateOffset:i_date, :]            
            testing_data = data.loc[i_date:i_date+retraining_period_dateOffset, :]
            
            if len(training_data) > 0 and len(testing_data) > 0:
                # Append train and retrain periods for chart rendering
                training_dates.append([training_data.index[0].timestamp() * 1000, training_data.index[-1].timestamp() * 1000])
                retraining_dates.append([testing_data.index[0].timestamp()* 1000, testing_data.index[-1].timestamp() * 1000])
            
            if ignore_outliers == True:
                # select training data
                # find outliers and remove them from the trainings data
                outliers_index = find_outliers_extended(threshold_instance.dependent_variable, training_data, bins=10, minRelModeWeight=0.1, min_peak_height=None)
                if np.sum(outliers_index) > 0:
                    outliers_df_list.append(training_data.loc[outliers_index])
                    training_data = training_data.loc[~outliers_index, :]

            # TODO simplify loop with isin?
            #44740 Remove duplicate indexes 
            for idx in testing_data.index.tolist():
                if idx in previous_testing_data:
                    testing_data = testing_data.drop(idx)
            previous_testing_data = testing_data.index.tolist()            

            try:               
                threshold_instance.fit(training_data)
            except ValueError as ve:
                errors = errors + "\n" + "".join(traceback.TracebackException.from_exception(ve).format())
                continue
            try:
                transf_df = threshold_instance.transform(testing_data)
            except ValueError as ve:
                errors = errors + "\n" + "".join(traceback.TracebackException.from_exception(ve).format())
                continue            
            df_list.append(transf_df)

        if len(df_list) > 0:
            results_df = pd.concat(df_list)
            # Merge outliers and remove duplicates
            results_df['outlier'] = False
            if len(outliers_df_list) > 0:
                outliers_df = pd.concat(outliers_df_list)
                outliers_df = outliers_df.loc[ ~outliers_df.index.duplicated(), : ] # remove all duplicated outliers
                for idx in outliers_df.index:
                    if idx in results_df.index:
                        results_df.loc[idx, 'outlier'] = True

        return results_df, training_dates, retraining_dates

    def parse_config_model(self, config_model: dict):
        self.INTERVAL = config_model['interval']
        self.TRAINING_DATES = config_model['trainingDates']
        self.TRAINING_PERIOD = config_model['periodsDay']['train']
        self.RETRAINING_PERIOD = config_model['periodsDay']['retrain']
        self.IGNORE_EVENTS = config_model['periodsDay']['ignoreEvents'] # Allows to ignore events in training data
        self.SENSITIVITY_UPPER_THRESHOLD = config_model['sensitivityThreshold']['upper']
        self.SENSITIVITY_LOWER_THRESHOLD = config_model['sensitivityThreshold']['lower']
        self.EVENT_COUNT_UPPER = config_model['eventCount']['upper']
        self.EVENT_COUNT_LOWER = config_model['eventCount']['lower']
        self.MIN_ERROR_DIST_UPPER = config_model['minErrorDistance']['upper']
        self.MIN_ERROR_DIST_LOWER = config_model['minErrorDistance']['lower']
        self.THRESHOLD_COLOR = config_model['thresholdColor']
        self.NODES = config_model['nodes']
        self.MIN_EL_PER_BIN = config_model['minElementsPerBin'] if 'minElementsPerBin' in config_model else None # GroupingThreshols only: Defines the minimum amount of group items
        self.DEPENDENT_VARIABLES = config_model['dependentVariables']  
        self.INDEPENDENT_VARIABLES = config_model['independentVariables']  

    def create_thresold_instance(self, method:str, dependent_variables:list, explanatory_vars:list[str], model_name:str = None, model_params = {}):
        threshold_class = METHODS_CLASS_DICT[method]
        q_lower = self.SENSITIVITY_LOWER_THRESHOLD / 100.0
        q_upper = 1.0 - self.SENSITIVITY_UPPER_THRESHOLD / 100.0
        
        model = None
        if model_name:
            model = self._regressor_model_list[model_name](**model_params)                            
            
        return threshold_class(dependent_variables=dependent_variables, explanatory_vars=explanatory_vars, q_lower=q_lower, q_upper=q_upper, min_no_elements_per_bin=self.MIN_EL_PER_BIN, model=model)
    
    def do_backtesting(self, data_df:pd.DataFrame, threshold_instance):
        results_df, training_dates, retraining_dates = self.train_rolling_data(threshold_instance, data_df, self.TRAINING_PERIOD, self.RETRAINING_PERIOD, self.IGNORE_EVENTS)
        if len(results_df) == 0:
            raise_bai_exception('tooFewDataThresholdCalc')
        
        outputs = {
            'thresholds': results_df,
            'trainingDates': training_dates,
            'reTrainingDates': retraining_dates,
        }
        
        return outputs    
    

    def bake_thresholds(self, thresholds_df:pd.DataFrame):
        # TODO: Quick fix. try to set None on init instead    
        thresholds_df.replace([np.inf, -np.inf, np.nan], None, inplace=True)
        
        # Add index as a simple column named x for storing purpose
        # thresholds_df['x'] = thresholds_df.index   
        # thresholds_df['x'] = thresholds_df.index.strftime(DATETIME_FORMATS['local'])        
        # thresholds_df['ts'] = thresholds_df['dateTime'].astype(np.int64) // 10 ** 6    

        return thresholds_df
        
    
    def do_single_train(self, threshold_instance, train_df:pd.DataFrame, transform_df:pd.DataFrame):      
        results_df = pd.DataFrame()        
        try:
            threshold_instance.fit(train_df)
            results_df = threshold_instance.transform(transform_df)            
        except ValueError as ve:
            # todo log traceback with traceback.format_exc()
            raise_bai_exception(ve.args[0])            
                        
        outputs = {
            'thresholds': results_df,
            'events': {}
        }        
        
        return outputs, threshold_instance            
        
    def do_single_transform(self, data:pd.DataFrame, explanatory_vars:list, config_model:dict) -> tuple[pd.DataFrame, any]:
        self.parse_config_model(config_model)
        
        model_params = get_model_params(config_model)
        threshold_instance = self.create_thresold_instance(config_model['task']['algorithm']['method'], config_model['dependentVariables'], explanatory_vars, config_model['task']['algorithm']['model'], model_params)
        results_df = threshold_instance.transform(data)
        return results_df, threshold_instance

    def set_alarm_flags(self, data:pd.DataFrame, skip_first_rows = False):
        """Raise alarms based on too many events

        :param data pd.DataFrame containing events, i.e. with column 'event'
        :param lookback_horizon  window in the past to check for events (CAVEAT: given in number of samples *not* Delta_time)
        :return:
        """
        # Fix for rolling window functionality
        no_upper_events = self.EVENT_COUNT_UPPER + 1
        no_lower_events = self.EVENT_COUNT_LOWER + 1

        # Todo refactor rolling method?
        # as time passes, this could cost some time
        adata = data[['event_upper']].copy()
        event_counts_upper = adata.rolling(no_upper_events).sum()        
        if skip_first_rows:
            # Skip first self.EVENT_COUNT_UPPER rows otherwise the get nan values            
            data.iloc[self.EVENT_COUNT_UPPER:, data.columns.get_loc('sum_upper')] = event_counts_upper[self.EVENT_COUNT_UPPER:]
        else:
            data['sum_upper'] = event_counts_upper            
            data['alarm_upper'] = (data['sum_upper'] >= no_upper_events) & (data['event_upper'] == True)

        adata = data[['event_lower']].copy()
        event_counts_lower = adata.rolling(no_lower_events).sum()        
        if skip_first_rows:            
            # Skip first self.EVENT_COUNT_LOWER rows otherwise the get nan values        
            data.iloc[self.EVENT_COUNT_LOWER:, data.columns.get_loc('sum_lower')] = event_counts_lower[self.EVENT_COUNT_LOWER:]
        else:
            data['sum_lower'] = event_counts_lower                        
        data['alarm_lower'] = (data['sum_lower'] >= no_lower_events) & (data['event_lower'] == True)            
        
        # Clean events
        data['event_upper'][data['alarm_upper'] == True] = False
        data['event_lower'][data['alarm_lower'] == True] = False
        
        return data

    def set_event_flags(self, results_df:pd.DataFrame) -> dict:
        """Set events flags""" 
        idx_upper = results_df['dist_upper'] > self.MIN_ERROR_DIST_UPPER
        idx_lower = results_df['dist_lower'] > self.MIN_ERROR_DIST_LOWER
        results_df['event_upper'] = idx_upper
        results_df['event_lower'] = idx_lower

        return results_df