import os, sys
import argparse

import warnings
warnings.filterwarnings("ignore")

currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)

# Append scripts folder
parentdir = os.path.dirname(parentdir)
sys.path.append(parentdir)


import pandas as pd
from metrics.metric_utils import get_node_name, get_temporal_var_list, get_feature_template, get_node_unique_identifier
import json
from sklearn.ensemble import RandomForestRegressor
import traceback

from intelli_utilities import BaiException


import logging
logging.basicConfig(level=logging.ERROR, format='%(message)s', stream=sys.stderr)

def get_important_features(y:list, X:list, independent_vars:list[dict]) -> pd.DataFrame:
    """Train randomForest and return most important features by given threshold"""
    rfr = RandomForestRegressor(n_jobs=-1)
    rfr.fit(X, y)
    
    df = pd.DataFrame(independent_vars)
    df['importance'] = rfr.feature_importances_
    df = df[df['importance'] >= 0]
    df.sort_values(by='importance', ascending=False, inplace=True)
    df.fillna('', inplace=True)
    
    # Todo integrated xgb feature importance
    # xgb_model = xgb.XGBRegressor()
    # xgb_model.fit(X, y)    
    
    # xgb_feature_importances = pd.Series(xgb_model.feature_importances_, index=col_names)
    # xgb_feature_importances = xgb_feature_importances.loc[lambda x : x > threshold]    
    
    # feature_extension = [ xgb_feature for xgb_feature in xgb_feature_importances.index if xgb_feature not in feature_importances.index]
    # feature_extension = feature_importances.concat(feature_extension)

    return df


def pick_independent_nodes(nodes:list, dependent_node:dict) -> list:
    independent_nodes = []
    for node in nodes:
        node_name = get_node_unique_identifier(node['pmModel'])
        dependent_node_name = get_node_unique_identifier(dependent_node['pmModel'])
        if node_name != dependent_node_name:
            feature = get_feature_template(get_node_name(node), node['uuid'], node['pmModel'])
            independent_nodes.append(feature)
    return independent_nodes


def main(args):
    response = {}
    input_uuid = args.inputDataUuid

    with open('spool/{}/input.json'.format(input_uuid), 'r') as file:
        input_data = json.load(file)

    nodes = input_data['nodes']
    interval = input_data['interval']

    if not input_data['dmsDataFrame']:
        raise BaiException('noDataFound')

    data_df = pd.DataFrame(input_data['dmsDataFrame'])
    data_df = data_df.set_index('dateTime', drop=False)
    data_df.index = pd.to_datetime(data_df.index)      
    data_df.dropna(inplace=True)

    dependent_nodes = []
    for node in nodes:
        if node['dependentVariable']:
            feature = get_feature_template(get_node_name(node), node['uuid'], node['pmModel'])
            dependent_nodes.append(feature)

    independent_var_nodes = get_temporal_var_list(interval)
    if 'holiday' in data_df.columns:
        independent_var_nodes.append(get_feature_template('holiday'))

    """Creates sets on important features for each dependent variable"""
    important_feature_sets = []
    for dependent_node in dependent_nodes:                
        dependent_node_name = get_node_unique_identifier(dependent_node['pmModel'])
        independent_nodes = pick_independent_nodes(nodes, dependent_node) + independent_var_nodes
        
        # merge temporal var and independent node names
        independent_col_names = []
        for node in independent_nodes:
            col_name = node['name']
            if node['pmModel']['path'] != '':
                col_name = get_node_unique_identifier(node['pmModel'])
            independent_col_names.append(col_name)
            
        y = data_df[dependent_node_name]        
        X = data_df[independent_col_names]
        important_features_df = get_important_features(y.tolist(), X.values, independent_nodes)
                
        important_feature_sets.append({
                'dependentVar': dependent_node,
                'independentVars': important_features_df.to_dict('records'),
            }
        )
    response = json.dumps(important_feature_sets)              
    print(response)
    
if __name__ == "__main__":
    try:
        parser = argparse.ArgumentParser()
        parser.add_argument('--inputDataUuid', help='Uuid of the node to load')
        parser.add_argument('--token', help='Security token for services')
        parser.add_argument('--processManager', help='Base url of the process manager')
        args = parser.parse_args()
        main(args)
    except BaiException as be:
        error = "{}".format(be)
        sys.stderr.write(error)  
    except Exception as e:
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]        
        error = "{} {} {} {} {}".format(e, exc_type, fname, exc_tb.tb_lineno, traceback.format_exc())
        sys.stderr.write(error)        
