import yaml

import requests
import itertools

def get_feature_template(name:str, uuid = '', pmModel = None):
    if pmModel == None:
        pmModel = {
        'instanceName': '',
        'instanceUuid': '',
        'path': '',
        'processManagerUuid': '',
        'processUuid': '',
    }

    return {
    'uuid': uuid,
    'name': name,
    'pmModel': pmModel
}

def print_stream(msg):
    print(msg, flush=True)

def fetch_job(job_uuid:str, token:str, process_manager:str):
    response = requests.get("{}/service/rest/v1/filedb/app/BAI/jobs/{}.yaml".format(process_manager, job_uuid), headers={'x-access-token': token}, verify=False)
    response.raise_for_status()    
    file_db = yaml.safe_load(response.content)
    return file_db['data']['data']

def get_node_unique_identifier(pmModel:dict):
    return '{}_{}'.format(pmModel['processUuid'], pmModel['path'])

def get_colname_list(nodes:list):
    explanatory_var_name_list = []
    for node in nodes:
        if node['uuid'] != '':
            explanatory_var_name_list.append(get_node_unique_identifier(node['pmModel']))
        else:
            explanatory_var_name_list.append(node['name']) 
            
    return explanatory_var_name_list            

def get_node_name(node:dict) -> str:
    node_name = ''
    if 'name' in node['extInfos']:
        node_name = node['extInfos']['name']

    if 'unit' in node['extInfos'] and node['extInfos']['unit'] != '':
        pattern = '[{}]'
        if node_name != '':
            pattern = '{} [{}]'            
        node_name = pattern.format(node_name, node['extInfos']['unit'])
    return node_name

def get_activated_features(features:list, mandatory:bool):
    activated_features = []    
    for var in features:
        if var['mandatory'] == mandatory and var['active'] == True:
            feature = get_feature_template(var['name'], var['uuid'], var['pmModel'])
            activated_features.append(feature)
    return activated_features

def get_dependent_independent_vars_combinations(important_features:list[dict]):
    """
    @param nodes: list of nodes
    @param explanatory_temporal_vars: List of temporal var names format [{name: 'month'}, {name: 'woy'}]
    @return combinations:
    [
        [{'dependent_var': {'name': 'Heat [LUG]', 'uuid': '2ee875db-2089-42e7-8426-36746cf77ca6'}],
        [{'independent_vars': [
            [{'name': 'woy'}], 
            [{'name': 'month'}, {'name': 'woy'}],
            ...
        ]
    ]        
    """    
    combinations = []                    
    for important_feature_set in important_features: 
        dependent_var = important_feature_set['dependentVar']
        
        independent_uuids_mandatory = get_activated_features(important_feature_set['independentVars'], True)
        independent_uuids_optional = get_activated_features(important_feature_set['independentVars'], False)
        
        independent_var_combinations = create_independent_vars_combination(independent_uuids_optional, independent_uuids_mandatory)                              
        combinations.append({
            'dependent_var': dependent_var, 
            'independent_vars': independent_var_combinations
        })
    return combinations   

def create_independent_vars_combination(explanatory_optional_nodes:list, explanatory_mandatory_nodes:list, additional_explanatory_vars:list = []) -> list:
    """
    @params explanatory_nodes: list of explanatory nodes
    @params expl_vars_name_list: List of all explanatory col names. This are not originaal but like explanatory_var_x
    @params additional_explanatory_vars: list of additional column names to consider for combinations
    
    Returns a list with all explanatory variable combinations with the respective uuid for explanatory nodes
    [
        [{'name': 'Aussenfühler °C', 'uuid': 'e8920a3b-8204-422f-b1be-29f0004a50be', 'path':'BN028NA:Netatmo:Aussen:Temperature'}, {'name': 'hod'}],
        [{'name': 'Aussenfühler °C', 'uuid': 'e8920a3b-8204-422f-b1be-29f0004a50be', 'path':'BN028NA:Netatmo:Aussen:Temperature'}]
    ]        
    """        
    # Prepend additional variables
    explanatory_optional_vars = additional_explanatory_vars + explanatory_optional_nodes

    # Create optional explanatory combinations
    explanatory_combinations = []
    for i in range(len(explanatory_optional_vars)):
        r = i + 1
        explanatory_combinations = explanatory_combinations + list(itertools.combinations(explanatory_optional_vars, r))

    # Prepend mandatory explanatory variables
    explanatory_combinations = [explanatory_mandatory_nodes + list(i) for i in explanatory_combinations]

    # Append mandatory only variables
    if explanatory_mandatory_nodes:
        explanatory_combinations.append(explanatory_mandatory_nodes)
    
    return explanatory_combinations


def get_temporal_var_list(interval:int) -> list:
    
    #Add uuid property for data consistency
    explanatory_temporal_vars = [
        get_feature_template('woy'),
        get_feature_template('month'),
        get_feature_template('quarter'),
        get_feature_template('Mo2Sa-So'),
        get_feature_template('Mo2Fr-Sa-So'),
        get_feature_template('Mo2Fr-Sa2So'),
    ]

    # ----- Set time variable according to resolution
    match interval:
        case 3600:
            explanatory_temporal_vars.append(get_feature_template('hod'))
            explanatory_temporal_vars.append(get_feature_template('dow'))
        case -1:
            explanatory_temporal_vars.append(get_feature_template('dow'))
        case _:
            pass

    return explanatory_temporal_vars


def set_mandatory_flag(important_feature_sets, nodes):
    for set in important_feature_sets:
        for independent_node in set['independentVars']:
            if independent_node['uuid'] != '':
                for node in nodes:
                    if node['uuid'] == independent_node['uuid']:
                        independent_node['mandatory'] = node['mandatory']
            else:
                independent_node['mandatory'] = False
    return important_feature_sets