from functools import reduce
import numpy as np
import pandas as pd
from typing import Union
from scipy.signal import find_peaks

##################################################################################
# Outlier detection
    
def _get_thresholds_of_mode(idx_maximum:int, histogram:np.ndarray, bin_edges:np.ndarray, minRelModeWeight:Union[float,None]=None, verbose=False) -> tuple:    
    """Given an index and the corresponding histogram (bins, edges), this function returns the edges 
    such that (i) no no empty bin lies within them and (ii) the idx_maximum lies also between them. 
    
    :param int idx_maximum: index of the maximum within a mode
    :param np.ndarray histogram: histogram to be analyzed
    :param np.ndarray bin_edges: edges of the histogram to be analyzed
    :param Union[float,None] minRelModeWeight: Relative weight of the mode, so edges will be calculated. If the counts in the mode
        devided through all counts in the histogram are smaller than the provided threshold, returned bin edges are
        the min and max of the 'bin_edges'. This condition will be ignored if set to None (default).
    :param bool verbose: if outputs should be printed 
    :return: tupel containing loer and upper threhold
    """
    total_counts = np.sum(histogram)
    idx_noCounts = np.nonzero(histogram == 0)[0]
    # get lower threshold
    smaller_indices = np.nonzero((idx_noCounts < idx_maximum))[0]
    idx_lower_threshold = 0 if len(smaller_indices) == 0 else idx_noCounts[smaller_indices[-1]]+1 # if more than one index -> take largest index below the max
    # get upper threshold
    larger_indices = np.nonzero((idx_noCounts > idx_maximum))[0]
    idx_upper_threshold = len(histogram) if len(larger_indices) == 0 else idx_noCounts[larger_indices[0]]  # if more than one index -> take smalles index above the max
    if not (minRelModeWeight == None):
        cim = np.sum(histogram[idx_lower_threshold: idx_upper_threshold]) # integral count of mode
        if verbose: print(f"Counts in mode - abs: {cim}, rel: {cim/total_counts:0.3f}")
        if cim/total_counts < minRelModeWeight: # mode is not relevant if its relative weight is smaller than threshold minRelModeWeight
            idx_lower_threshold, idx_upper_threshold = -1, 0 
    lower_threshold = bin_edges[idx_lower_threshold]
    upper_threshold = bin_edges[idx_upper_threshold]
    if verbose: print(f"Thresholds -> lower idx: {idx_lower_threshold}, value: {lower_threshold}, upper idx: {idx_upper_threshold}, value: {upper_threshold}")
    return lower_threshold, upper_threshold
    
def find_outliers_extended(dependent_variable:str, data:pd.DataFrame, bins:int=10, min_peak_height:Union[int,None]=None, 
                           minRelModeWeight:Union[float,None]=None, verbose=False, **kwargs) -> np.ndarray:
    """Returns a boolean array with `True` at the position of the outliers in the column dependent_variable of provided data.
    
    Approach works by buildingwith a histogram of the data. It then searches for the maxima
    of the distribution and for each maxima looks for the first bin with count==0 on both sides. 
    Outliers are those points that are for all maxima beyond the first bin with count 0
    :param pd.DataFrame data: data frame with column dependent_variable that is investigated for outliers
    :param int bins: number of bins to be used for binning the data
    :param Union[int,None] min_peak_height: Parameter for peak finding algorithm. Minimal number of counts 
        a maximum has to have in order to be considered as a peak of a mode. Defaults to `None` in wich case
        no threshold is used.
    :param Union[float,None] minRelModeWeight: Relative weight a mode (i.e. sum over all counts beteen two
        bins with count eequal 0 devided through total number of counts) is required to have so it is not 
        considered an outlier. Defaults to `None` in which case the condition will be ignored. 
    return np.ndarray: boolean index array (of length data.shape[0]) with `True` for outliers
    """
    # create histogram
    dataNoNan = data.copy().dropna(axis=0, subset=[dependent_variable])
    hist, bin_edges = np.histogram(dataNoNan[dependent_variable], bins=bins)
    
    if verbose:
        print(f"\nfind_outliers_extended:")
        print(f"hist     : {hist}")
        print(f"bin edges: {bin_edges}")
    # find peaks
    ext_hist = np.concatenate(([0], hist, [0])) # extend histogram with count=0 at each edge to also detect peaks at the edge
    peak_indices, properties = find_peaks(ext_hist, height=min_peak_height)
    peak_indices = peak_indices - 1 # backtransform indices from ext_hist -> hist
    # find outliers
    idx_noCounts = np.nonzero(hist == 0)[0]
    if not np.any(idx_noCounts): # check that we have bins without a count 
        return np.full(data.shape[0], False, dtype=np.bool_) 
    else:
        isOutlier_list = []
        isOutlier = np.array([])
        lower_threshold, upper_threshold = 0, 0  # initialize thresholds
        for idx_maximum in peak_indices:
            if (lower_threshold <= bin_edges[idx_maximum]) and (bin_edges[idx_maximum] < upper_threshold): 
                if verbose: print(f"peak index: {idx_maximum} already treated")
                continue # skip peaks in already treated modes
            if verbose: print(f"peak index: {idx_maximum}")
            lower_threshold, upper_threshold = _get_thresholds_of_mode(idx_maximum, hist, bin_edges, minRelModeWeight, verbose=verbose)
            ol = (data[dependent_variable].values < lower_threshold) |  (data[dependent_variable].values > upper_threshold)
            isOutlier_list.append( ol )

        if len(isOutlier_list) > 0:
            isOutlier = reduce((lambda x, y: x & y), isOutlier_list)

        return isOutlier

