import pandas as pd
import numpy as np
from collections import deque
from utils import *

class TimeSlotMonitor():
    def __init__(self,
                 target_col,
                 lookback=7,
                 th_percent=0.5,
                 min_th=0.5,
                 alpha=0.2,
                 monitoring_time_start='22:00',
                 monitoring_time_end='05:00',
                 daily_alarm_count=3,
                 weekly_alarm_count=3,
                 sunday_alarm_count=2,
                 max_power_alarm = None,
                 verbose=False) -> None:


        self.target_col = target_col
        self.lookback = lookback
        self.th_percent = th_percent
        self.min_th = min_th
        self.alpha = alpha
        self.monitoring_time_start = monitoring_time_start
        self.monitoring_time_end = monitoring_time_end
        self.oneDay = 24 # one day in hours. Change for other resolution
        self.daily_alarm_count = daily_alarm_count
        self.weekly_alarm_count = weekly_alarm_count
        self.sunday_alarm_count = sunday_alarm_count
        self.max_power_alarm = max_power_alarm
        self.verbose = verbose

    def fit(self, X:pd.DataFrame):
        # get initial monitoring time consumption using the available data
        self.monitoring_times = deque()
        for idx, group in X.groupby('day_of_year'):
            if len(group) == self.oneDay:
                self.monitoring_times.append(group[self.target_col].between_time(self.monitoring_time_start, self.monitoring_time_end).mean())

        if self.verbose:
            monitoring_time_mean = np.mean(self.monitoring_times)
            monitoring_time_p95 = np.percentile(self.monitoring_times, q=95)

            print('Estimated consumption using training data:')
            print('\t Mean Monitoring time: {:.3f} (N={})'.format(monitoring_time_mean, len(self.monitoring_times)))
            print('\t P95-Monitoring time: {:.3f} (N={})'.format(monitoring_time_p95, len(self.monitoring_times)))

        # calculate threshold
        self.curr_th = calculate_threshold(monitoring_time=self.monitoring_times, th_percent=self.th_percent, min_th=self.min_th)

    def predict(self, X:pd.DataFrame):

        data = []
        tmp = list(self.monitoring_times)[-self.lookback:]
        self.monitoring_times = deque(maxlen=self.lookback)
        self.monitoring_times.extend(tmp)

        self.daily_alarms = deque(maxlen=self.daily_alarm_count)
         # keys = day_of_week, values = deque with events
        self.weekly_alarms = dict()
        for day_of_week in range(0,7):
            self.weekly_alarms[day_of_week] = deque(maxlen=self.weekly_alarm_count)
        self.sunday_alarms = deque(maxlen=self.sunday_alarm_count)

        for idx, group in X.groupby('day_of_year'):
            if len(group) == self.oneDay:
                timestamp = group.index[0]
                # calculate montoring time mean consumption
                curr_monitoring_time_mean = group[self.target_col].between_time(self.monitoring_time_start, self.monitoring_time_end).mean()
                # detect event
                event = (curr_monitoring_time_mean > self.curr_th).astype(int)

                # detect alarms
                self.daily_alarms.append(event)
                self.weekly_alarms[timestamp.day_of_week].append(event)
                if timestamp.day_of_week == 6:
                    self.sunday_alarms.append(event)
                alarm = 0
                if event:
                    # check daily alarms
                    # mark alarms via prime factor
                    if sum(self.daily_alarms) == self.daily_alarm_count:
                        alarm = alarm + 1
                    if sum(self.weekly_alarms[timestamp.day_of_week]) == self.weekly_alarm_count:
                        alarm = alarm + 3
                    if sum(self.sunday_alarms) == self.sunday_alarm_count:
                        alarm = alarm + 5
                    # if self.max_power_alarm is not None:
                    #     if curr_day_mean > self.max_power_alarm:
                    #         alarm = alarm + 7

                # save current state
                data.append({'dateTime': pd.to_datetime("{} 00:00:00".format(idx)), 'time_slot_mean': curr_monitoring_time_mean, 'th': self.curr_th, 'event': event, 'alarm': alarm})

                # adapt threshold based on exponential moving average of consumption data
                self.monitoring_times.append(curr_monitoring_time_mean)

                new_th = calculate_threshold(monitoring_time=self.monitoring_times, th_percent=self.th_percent, min_th=self.min_th)
                self.curr_th = exponential_moving_average(new_th, self.curr_th, self.alpha)

        return pd.DataFrame(data)