// Call ngui function to create a ui config.
import useComponent from '/service/rest/v1/files/data/public/js/useComponent.js'
const components = useComponent()

import useProcessManager from '/service/rest/v1/files/data/public/js/useProcessManager.js'
const processManager = useProcessManager()

window.ngui(() => {
  return {
    mixins: [
      processManager.mixin(),
    ],
    beforeCreate() {
      components.setParent(this)
      processManager.setParent(this)
    },

    // Listen on unmounted event...
    unmounted() {
      // logic
      this.reset()
      this.isUnmounted = true
    },

    mounted() {
      // logic
      this.reset()
      this.isUnmounted = false
    },
    
    // The returned object can be accessed with this within any method...
    // Do not use this.ctx within this function because it is not yet created.
    data() {
      return {
        // VAR
        isUnmounted: false,
        loading: false,
        dialog: {
          edit: false,
          deleteProject: false,
          deleteProcessManager: false,
          instance: false,
          goToNewProject: false,
        },
        processManagerUuid: this.$route.query.service,
        intervalId: null,
        intervalTime: 1000,
        selectable: false,
        newInstanceUuid: '',
        chartOptions: {
          chart: {
            type: 'line',
          },
          xaxis: {
            type: 'datetime',
          },
          yaxis: [
            {
              id: 'cpu',
              min: 0,
              max: 100,
              startOnTick: true,
              endOnTick: true,
              labels: {
                style: {
                  color: 'red'
                }
              }
            },
            {
              id: 'memory',
              min: null,
              max: null,
              startOnTick: true,
              endOnTick: true,
              labels: {
                style: {
                  color: 'blue'
                }
              }
            }
          ],
          tooltip: {
              enabled: true,
              decimals: 2,
          },
        },
        timeSeries: [
          {
              name: 'CPU [%]',
              id: 'cpu',
              data: [],
              yAxis: 'cpu',
              color: 'red'
          },
          {
              name: 'Memory [MB]',
              id: 'memory',
              data: [],
              yAxis: 'memory',
              color: 'blue'
          }
        ],
        playTimeseries: true,
        instancesSettingsObject: {},
        monitoring: undefined,
        pingInterval: 0,
      }
    },

    watch: {
      instancesStatus(status) {
        this.handleTimeSeries(status)
      }
    },

    // Define methods for the ui...
    methods: {
      // functions
      async init() {
        try {
          this.loading = true;
          this.processManager = await this.ctx.processManager.fetchManager(this.processManagerUuid)
          this.instancesSettings = await this.ctx.processManager.fetchInstancesSettings(this.processManagerUuid)
          this.pingStart()

          this.instancesSettings.forEach((instance) => {
            this.instancesSettingsObject[instance.uuid] = instance
          })

          this.pollStatus()
        } catch (e) {
          this.ctx.notify.notifyApiError(e)
        }
        this.loading = false
      },
      
      reset() {
        this.stopPoll()
        this.pingStop()
      },
      
      fetchMonitoring() {
        if (this.monitoring && this.monitoring.options.error) {
          this.monitoring = undefined
        }

        this.ctx.httpClient.get(`/monitoring/${this.processManagerUuid}?timeout=1`).then((res) => {
          this.monitoring = res.data
        })
      },

      pingStart() {
        this.pingStop()
        this.fetchMonitoring()
        this.pingInterval = window.setInterval(() => {
          this.fetchMonitoring()
        }, 30 * 1000)
      },

      pingStop() {
        window.clearInterval(this.pingInterval)
      },
    
      onDeleteSelected() {
        clearInterval(this.intervalId)
        this.removeSelectedInstances(this.processManagerUuid).then(() => this.pollStatus())
      },

      removeManager() {
        clearInterval(this.intervalId)
        this.ctx.processManager.deleteManager(this.processManagerUuid)
          .then(() => {
            this.$router.push('/process-manager')
          })
      },
  
      async pollStatus() {
        clearInterval(this.intervalId)
        this.instancesStatus = await this.ctx.processManager.fetchInstancesStatus(this.processManagerUuid)
        this.startPollInterval()
      },
  
      startPollInterval() {
        clearInterval(this.intervalId)
        if (this.isUnmounted) {
          return
        }
        this.intervalId = setInterval(async () => {
          this.instancesStatus = await this.ctx.processManager.fetchInstancesStatus(this.processManagerUuid)
        }, this.intervalTime)
      },

      stopPoll() {
        clearInterval(this.intervalId)
      },

      getActionMenu(props, permissions) {
        return components.element('NgSection').on({ onClick: this.withModifiers(() => {}, ['stop'])}).append(() => [
          components.menu().attrs({ 
            items: processManager.getActionMenu(props.row, permissions), 
            enableGlobalClick: true, 
          }).on({ 
            'onClick:menu': (ident) => {
              if (ident === 'edit') {
                this.instanceForm = props.row
                this.dialog.instance = true
              } else if (ident === 'delete') {
                this.selectedInstances = [props.row]
                this.dialog.deleteProject = true
              } else {
                this.ctx.processManager.putInstanceAction(this.processManagerUuid, props.row.uuid, ident)
              }
            },
          })
        ])
      },

      getProcessesMenu(props) {
        const processes = !props.row.processes || props.row.processes.length <= 0 ? [{ name: 'noData' }] : props.row.processes
        return components.element('NgSection').on({ onClick: this.withModifiers(() => {}, ['stop'])}).append(() => [
          components.menu().attrs({
            permissions: this.ctx.globals.getPermissions(['processesRead']),
            items: processes.map((process) => ({
              label: process.name,
              id: process.name,
              disable: process.status !== 2 || !window.processManager.hasModule(process.name),
              icon: process.status !== 2 ? 'block' : 'done',
              click: () => {
                components.router().push(`/modules/${process.name}/index`, { 
                  service: this.$route.query.service,
                  instance: props.row.uuid,
                  process: process.uuid,
                })
              }
            })), 
            enableGlobalClick: false, 
          })
        ])
      },

      getTotalStatus(status) {
        let memUsage = 0
        let cpuUsage = 0

        status.forEach((stat) => {
          memUsage += stat.memUsage
          cpuUsage += stat.cpuUsage
        })
        
        return {
          memUsage,
          cpuUsage,
        }
      },
      
      handleTimeSeries(status) {
        if (!this.playTimeseries) {
          return
        }
        const timestamp = new Date().valueOf()
        const { memUsage, cpuUsage } = this.getTotalStatus(status)
        const cpuSerie = [timestamp, Number(cpuUsage.toFixed(1))]
        const memSerie = [timestamp, Number((memUsage / (1 * 1000 * 1000)).toFixed(0))]
        this.timeSeries[0].data.push(cpuSerie)
        this.timeSeries[1].data.push(memSerie)

        if (this.timeSeries[0].data.length > 600) {
          this.timeSeries[0].data.shift()
        }

        if (this.timeSeries[1].data.length > 600) {
          this.timeSeries[1].data.shift()
        }

         this.timeSeries = [... this.timeSeries];
      },

      handlePlayTimeseries() {
        this.playTimeseries = !this.playTimeseries
        if (this.playTimeseries) {
          this.timeSeries[0].data = []
          this.timeSeries[1].data = []
        }
      },

      getDropdown(props) {
        return components.td().append(() => [
          components.element('NgDropdown').on({ 
            onClick: this.withModifiers(() => {}, ['stop'])
          }).attrs({ 
            color: 'primary', 
            label: '',
            flat: true,
            dense: true,
            permissions: this.ctx.globals.getPermissions(['instancesRead']),
        }).append(() => [
            components.card().attrs({ style: { minWidth: '320px' }, disableToolbar: false, closable: true }).append(() => [
              components.section().append(() => [
                components.list('instance').append(() => [
                  components.item().append(() => [
                    components.column().append(() => [
                      components.text(props.row.name),
                      processManager.getStatusComponents(props.row, 'flex', true),
                      processManager.getActionButtons(props.row, (ident) => {
                        this.ctx.processManager.putInstanceAction(this.processManagerUuid, props.row.uuid, ident)
                      }).attrs({ 
                        id: 'instance-card-action',
                        permissions: this.ctx.globals.getPermissions(['instancesUpdate']),
                      }),
                      components.input('enableAndAutoStart', 'checkbox', this.instancesSettingsObject[props.row.uuid].enabled, (val) => {
                        this.instancesSettingsObject[props.row.uuid].enabled = val
                        this.ctx.processManager.putInstanceSettings(this.processManagerUuid, props.row.uuid, this.instancesSettingsObject[props.row.uuid])
                      }).attrs({
                         permissions: this.ctx.globals.getPermissions(['instancesUpdate']),
                      })
                    ])
                  ]),
                ]),
              ]),
              components.section().append(() => [
                components.list('processes').append(() => [
                  components.renderIf(props.row.processes.length <= 0, components.item('noData')),
                  ...props.row.processes.map((process, index) => {
                    return components.item().append(() => [
                      components.column().append(() => [
                        components.text(process.name),
                        processManager.getStatusComponents(process, 'flex', true),
                        processManager.getActionButtons(process, (ident) => {
                          this.ctx.processManager.putProcessAction(this.processManagerUuid, props.row.uuid, process.uuid, ident)
                        }).attrs({ 
                          id: 'process-action',
                          permissions: this.ctx.globals.getPermissions(['processesUpdate']),
                        }),
                        components.input('enableAndAutoStart', 'checkbox', this.instancesSettingsObject[props.row.uuid].processes[index].enabled, (val) => {
                          this.instancesSettingsObject[props.row.uuid].processes[index].enabled = val
                          this.ctx.processManager.putProcessSettings(
                            this.processManagerUuid, 
                            props.row.uuid, 
                            this.instancesSettingsObject[props.row.uuid].processes[index].uuid, 
                            this.instancesSettingsObject[props.row.uuid].processes[index],
                          )
                        }).attrs({
                           permissions: this.ctx.globals.getPermissions(['processesUpdate']),
                        })
                      ]), 
                    ])
                  })
                ])
              ])
            ])
          ])
        ])
      }
    },
      
    // Define elements within the ui configuration...
    nodes() {
      return [
        // any ui component node can be used...
        components.page([26], () => this.init(), () => this.reset()).append(() => [

          components.breadcrumbs(processManager.getPMBreadcrumbs(this.processManagerUuid)),
          
          components.bar('manager').createMenu('file', [
            {
              id: 'edit',
              label: 'edit',
              icon: 'edit',
              permissions: this.ctx.globals.getPermissions(['managersUpdate']),
              click: () => {
                this.dialog.edit = true
              }
            },
            {
              id: 'delete',
              label: 'delete',
              icon: 'delete',
              separator: true,
              permissions: this.ctx.globals.getPermissions(['managersDelete']),
              click: () => {
                this.dialog.deleteProcessManager = true
              }
            },
          ]).attrs({
            id: 'manager-toolbar',
          }),
          
          components.section().attrs({ maxHeight: '80vh' }).append(() => [
            components.grid().append(() => [
              components.cell(4).append(() => [
                components.list(this.processManager.options.name).attrs({ 
                  dense: false, 
                  flat: false 
                }).append(() => [
                  components.item().append('section', () => [
                    processManager.getMonitoringBadge(this.monitoring),
                  ]),
                  components.item('name', this.processManager.options.name),
                  components.item('description', this.processManager.options.description),
                  components.item('id', this.processManager.options.slug),
                  components.item('url', this.processManager.options.url),
                  components.expandItem('performance').append(() => [
                    components.chart('', this.chartOptions, this.timeSeries, 200),
                    components.button('', this.playTimeseries ? 'stop' : 'play_arrow').on({
                      onClick: () => this.handlePlayTimeseries()
                    }).attrs({
                      color: this.playTimeseries ? 'negative' : 'positive',
                      tooltip: 'chart',
                      flat: true,
                      dense: true,
                    }),
                  ]),
                ]).attrs({
                  id: 'manager-card',
                })
              ]),
              components.cell(8).append(() => [
                components.card().attrs({ flat: !this.ctx.platform.browser }).append(() => [
                  components.table(
                    'instances', 
                    [
                      {
                        i18nLabel: '',
                        align: 'left',
                        name: 'list',
                        field: 'list',
                      },
                      {
                        i18nLabel: 'name',
                        align: 'left',
                        name: 'name',
                        field: 'name',
                        sortable: true,
                      },
                      {
                        i18nLabel: 'description',
                        align: 'left',
                        name: 'description',
                        field: 'description',
                        sortable: true,
                        style: 'max-width: 250px;',
                        classes: 'ellipsis',
                      },
                      {
                        i18nLabel: 'portOffsetPrefix',
                        align: 'left',
                        name: 'portOffsetPrefix',
                        field: 'portOffsetPrefix',
                        format: (portOffsetPrefix) => this.ctx.i18n.i18n.t('displayPortOffsetPrefix', { portOffsetPrefix }),
                        sortable: true,
                      },
                      {
                        i18nLabel: 'status',
                        align: 'left',
                        name: 'status',
                        field: 'status',
                        sortable: true,
                      },
                      {
                        i18nLabel: 'action',
                        align: 'left',
                        name: 'action',
                        field: 'action'
                      },
                      {
                        i18nLabel: 'processes',
                        align: 'left',
                        name: 'processes',
                        field: 'processes'
                      }
                    ], 
                    this.instancesStatus,
                  ).attrs({
                    id: 'instances-table',
                    selection: 'multiple',
                    selected: this.selectedInstances,
                    rowKey: 'uuid',
                    grid: !this.ctx.platform.browser,
                    flat: true,
                    loading: this.loading,
                    loadingLabel: 'fetchInstances',
                    enableAddBtn: true,
                    enableDeleteBtn: true,
                    permissionsAdd: [25],
                    permissionsDelete: [28],
                    permissionsSelect: [28],
                    'onUpdate:selected': (val) => this.selectedInstances = val, 
                    'onClick:row': (props) => components.router().push('/process-manager/instances/index', { 
                      service: this.$route.query.service, 
                      instance: props.row.uuid,
                    }),
                    'onClick:add': () => {
                      this.instanceForm = {}
                      this.dialog.instance = true
                    },
                    'onConfirm:delete': () => {
                      this.onDeleteSelected()
                    },
                    maxHeight: this.ctx.platform.browser ? '80vh' : undefined,
                  }).append('body-cell-list', (props) => [
                    this.getDropdown(props)
                  ]).append('body-cell-name', (props) => [
                    components.td().append(() => [
                      components.flex().append(() => [
                        components.icon(
                          props.row.name === '_ProcessManager' ? 'home' : 'folder',
                          props.row.name === '_ProcessManager' ? 'primary' : 'grey',
                          'xs',
                        ),
                        components.text(props.value).attrs({ type: props.row.name === '_ProcessManager' ? 'bold' : '' })
                      ])
                    ])
                  ]).append('body-cell-status', (props) => [
                    components.td().append(() => [
                      processManager.getStatusBadge(props.row),
                    ])
                  ]).append('body-cell-action', (props) => [
                    components.td().append(() => [
                      this.getActionMenu(props, {
                        update: [27],
                        delete: [28]
                      }).attrs({
                        permissions: this.ctx.globals.getPermissions(['instancesUpdate', 'instancesDelete'])
                      })
                    ])
                  ]).append('body-cell-processes', (props) => [
                    components.td().append(() => [
                      this.getProcessesMenu(props).attrs({
                        permissions: this.ctx.globals.getPermissions(['processesRead'])
                      })
                    ])
                  ]).append('item', (props) => [
                    components.tdItem(components.card().append(() => [
                      components.renderIf(this.selectable, components.section().append(() => [
                        components.input('select', 'checkbox', props.selected, (val) => props.selected = val).attrs({
                          permissions: this.ctx.globals.getPermissions(['instancesDelete']),
                        }),
                      ])),
                      components.section().append(() => [
                        components.item(props.row.name, props.row.name !== props.row.description && props.row.description ? props.row.description : '').attrs({
                          clickable: true,
                          onClick: () => components.router().push('/process-manager/instances/index', { 
                            service: this.$route.query.service, 
                            instance: props.row.uuid,
                          }),
                        }).append('side', () => [
                          this.getActionMenu(props, {
                            update: [27],
                            delete: [28]
                          }),
                        ]),
                        components.item().append('section', () => [
                          processManager.getStatusBadge(props.row),
                        ]).append('side', () => [
                          this.getDropdown(props)
                        ]),
                      ]),
                    ]))
                  ])
                ]),
              ])
            ]),
          ]),

          components.dialogDelete(this.selectedInstances.map((item) => item.name).join(', '), this.dialog.deleteProject, (val) => this.dialog.deleteProject = val).on({
            onOk: () => this.onDeleteSelected(),
            onCancel: () => this.selectedInstances = [],
          }), 

          components.dialogDelete(this.processManager.options.name, this.dialog.deleteProcessManager, (val) => this.dialog.deleteProcessManager = val).on({
            onOk: () => this.removeManager(),
          }), 

          components.formDialog('manager', this.dialog.edit, (val) => this.dialog.edit = val, 'public/forms/manager.js').formOn({
            onSubmit: (_, form) => {
              this.dialog.edit = false
              this.processManager = form
            },
            onCancel: () => this.dialog.add = false,
          }).formAttrs({
            slug: this.processManagerUuid,
          }),
          
          components.formDialog('instance', this.dialog.instance, (val) => this.dialog.instance = val, 'public/forms/instance.js').formOn({
            onSubmit: (res, _, isNew) => {
              this.dialog.instance = false
              if (isNew) {
                this.newInstanceUuid = res.instanceUuid,
                this.dialog.goToNewProject = true 
              }
            },
            onCancel: () => this.dialog.instance = false,
          }).formAttrs({
            instanceUuid: this.instanceForm ? this.instanceForm.uuid : '',
            withDMS: true,
          }).on({
            onHide: () => {
              this.instanceForm = {}
            }
          }),
          
          components.dialogConfirm('configureNewInstance', this.dialog.goToNewProject, (val) => this.dialog.goToNewProject = val).on({
            onOk: () => components.router().push('/process-manager/instances/index', { 
              service: this.$route.query.service, 
              instance: this.newInstanceUuid,
            }),
          }), 
        ]),
      ]
    },
  }
})