// Call ngui function to create a ui config.
import useComponent from '/service/rest/v1/files/data/public/js/useComponent.js'
const components = useComponent()

import useProcessManager from '/service/rest/v1/files/data/public/js/useProcessManager.js'
const processManager = useProcessManager()

window.ngui(() => {
  return {
    mixins: [
      processManager.mixin(),
    ],
    beforeCreate() {
      components.setParent(this)
      processManager.setParent(this)
    },

    // Listen on unmounted event...
    unmounted() {
      // logic
      this.isUnmounted = true
      this.stopPoll()
    },
    mounted() {
      // logic
      this.isUnmounted = false
      this.stopPoll()
    }, 
    // The returned object can be accessed with this within any method...
    // Do not use this.ctx within this function because it is not yet created.
    data() {
      return {
        // VAR
        isUnmounted: false,
        processManagerUuid: this.$route.query.service,
        instanceUuid: this.$route.query.instance,
        loading: false,
        dialog: {
          remove: false,
          addProcess: false,
          deleteProcesses: false,
          process: false,
          instance: false,
          deleteInstance: false,
        },
        chartOptions: {
          chart: {
            type: 'line',
          },
          xaxis: {
            type: 'datetime',
          },
          yaxis: [
            {
              id: 'cpu',
              min: 0,
              max: 100,
              startOnTick: true,
              endOnTick: true,
              labels: {
                style: {
                  color: 'red'
                }
              }
            },
            {
              id: 'memory',
              min: null,
              max: null,
              startOnTick: true,
              endOnTick: true,
              labels: {
                style: {
                  color: 'blue'
                }
              }
            }
          ],
          tooltip: {
              enabled: true,
              decimals: 2,
          },
        },
        timeSeries: [
          {
              name: 'CPU [%]',
              id: 'cpu',
              data: [],
              yAxis: 'cpu',
              color: 'red'
          },
          {
              name: 'Memory [MB]',
              id: 'memory',
              data: [],
              yAxis: 'memory',
              color: 'blue'
          }
        ],
        playTimeseries: true,
        intervalId: null,
        intervalTime: 1000,
        selectable: false,
        processSettingsObject: {}
      }
    },

    watch: {
      instanceStatus(status) {
        this.handleTimeSeries(status)
      }
    },

    // Define methods for the ui...
    methods: {
      // functions
      async init() {
        try {
          this.loading = true
          // Reset config service to local to get the correct manager, afterwards we add the active service to get a secure communication.
          this.templates = await this.ctx.processManager.fetchTemplates(this.processManagerUuid, this.instanceUuid)
          this.processManager = await this.ctx.processManager.fetchManager(this.processManagerUuid)
          this.instanceSettings = await this.ctx.processManager.fetchInstanceSettings(this.processManagerUuid, this.instanceUuid)
          this.instanceStatus = await this.ctx.processManager.fetchInstanceStatus(this.processManagerUuid, this.instanceUuid)
          this.servers = await this.ctx.processManager.fetchServers(this.processManagerUuid, 'Api-Gateway')
          const as = this.servers.instances.find((inst) => inst.processes.find((proc) => proc.name === 'Api-Gateway'))
          if (as && as.processes && as.processes[0] && as.processes[0].servers && as.processes[0].servers[0]) {
            const aserver = as.processes[0].servers[0]
            this.apigatewayEndpoint = `${aserver.location.scheme}://localhost:${aserver.location.port}`
          } else {
            this.apigatewayEndpoint = `${window.location.protocol}//localhost:${window.location.port}`
          }
          this.initServers()
          this.pollStatus()
          this.instanceSettings.processes.forEach((process) => {
            this.processSettingsObject[process.uuid] = process
          })
        } catch (e) {
          this.ctx.notify.notifyApiError(e)
        }
        this.loading = false
      },
  
      async pollStatus() {
        this.stopPoll()
        this.instanceStatus = await this.ctx.processManager.fetchInstanceStatus(this.processManagerUuid, this.instanceUuid)
        this.startPollInterval()
      },

      async initServers() {
        const res = await this.ctx.processManager.fetchServersByInstance(this.processManagerUuid, this.instanceUuid)
        if (res.instances && res.instances.length) {
          this.instanceServers = res.instances[0]
        } else {
          this.instanceServers = {
            processes: []
          }
        }
      },

      onDeleteSelected() {
        this.removeSelectedProcesses(this.processManagerUuid, this.instanceUuid).then(() => this.init())
      },

      removeInstance() {
        this.stopPoll()
        this.ctx.processManager.deleteInstance(this.processManagerUuid, this.instanceUuid)
          .then(() => {
            this.ctx.notify.notifyDeleted()
            this.$router.push({
              path: '/process-manager/index',
              query: {
                service: this.$route.query.service,
              }
            })
          })
      },
  
      startPollInterval() {
        this.stopPoll()
        if (this.isUnmounted) {
          return
        }
        this.intervalId = setInterval(async () => {
          this.instanceStatus = await this.ctx.processManager.fetchInstanceStatus(this.processManagerUuid, this.instanceUuid)
        }, this.intervalTime)
      },

      stopPoll() {
        clearInterval(this.intervalId)
      },

      getActionMenu(props, permissions) {
        return components.element('NgSection').on({ onClick: this.withModifiers(() => {}, ['stop'])}).append(() => [
          components.menu().attrs({ 
            items: processManager.getActionMenu(props.row, permissions), 
            enableGlobalClick: true, 
          }).on({ 
            'onClick:menu': (ident) => {
              if (ident === 'edit') {
                this.processSettings = props.row
                this.dialog.process = true
              } else if (ident === 'delete') {
                this.selectedProcesses = [props.row]
                this.dialog.deleteProcesses = true
              } else {
                this.ctx.processManager.putProcessAction(this.processManagerUuid, this.instanceUuid, props.row.uuid, ident)
              }
            },
          })
        ])
      },

      getInstanceActionMenu() {
        return components.element('NgSection').on({ onClick: this.withModifiers(() => {}, ['stop'])}).append(() => [
          components.menu().attrs({ 
            items: processManager.getActionMenu(this.instanceStatus), 
            enableGlobalClick: true, 
          }).on({ 
            'onClick:menu': (ident) => {
              if (ident === 'edit') {
                this.instanceForm = this.instanceSettings
                this.dialog.instance = true
              } else if (ident === 'delete') {
                this.selectedProcesses = [this.instanceSettings]
                this.dialog.deleteInstance = true
              } else {
                this.ctx.processManager.putInstanceAction(this.processManagerUuid, this.instanceSettings.uuid, ident)
              }
            },
          })
        ])
      },

      handleTimeSeries(status) {
        if (!this.playTimeseries) {
          return
        }
        const timestamp = new Date().valueOf()
        const { memUsage, cpuUsage } = status
        this.timeSeries[0].data.push({
          x: timestamp,
          y: Number(cpuUsage.toFixed(1)),
        })
        this.timeSeries[1].data.push({
          x: timestamp,
          y:  Number((memUsage / (1 * 1000 * 1000)).toFixed(0)),
        })

        if (this.timeSeries[0].data.length > 600) {
          this.timeSeries[0].data.shift()
        }

        if (this.timeSeries[1].data.length > 600) {
          this.timeSeries[1].data.shift()
        }

        this.timeSeries = [... this.timeSeries];
        
      },

      handlePlayTimeseries() {
        this.playTimeseries = !this.playTimeseries
        if (this.playTimeseries) {
          this.timeSeries[0].data = []
          this.timeSeries[1].data = []
        }
      },

      getDropdown(props) {
        return components.td().append(() => [
          components.element('NgDropdown').on({ 
            onClick: this.withModifiers(() => {}, ['stop'])
          }).attrs({ 
            color: 'primary', 
            label: '',
            flat: true,
            dense: true,
            permissions: [31],
          }).append(() => [
            components.card().attrs({ style: { minWidth: '320px' }, disableToolbar: false, closable: true }).append(() => [
              components.section().append(() => [
                components.list('').append(() => [
                  components.item().append(() => [
                    components.column().append(() => [
                      components.text(props.row.name),
                      processManager.getStatusComponents(props.row, 'flex', true),
                      processManager.getActionButtons(props.row, (ident) => {
                        this.ctx.processManager.putProcessAction(this.processManagerUuid, this.instanceStatus.uuid, props.row.uuid, ident)
                      }).attrs({ 
                        id: 'process-action',
                      }),
                      components.input('enableAndAutoStart', 'checkbox', this.processSettingsObject[props.row.uuid].enabled, (val) => {
                        this.processSettingsObject[props.row.uuid].enabled = val
                        this.ctx.processManager.putProcessSettings(this.processManagerUuid, this.instanceUuid, props.row.uuid, this.processSettingsObject[props.row.uuid])
                      }).attrs({ 
                      }),
                    ]),
                  ])
                ])
              ])
            ])
          ])
        ])
      }
    },
      
    // Define elements within the ui configuration...
    nodes() {
      return [
        // any ui component node can be used...
        components.page([26], () => this.init(), () => this.stopPoll()).append(() => [

          components.breadcrumbs(processManager.getPMBreadcrumbs(this.processManagerUuid, this.instanceUuid)),
          
          components.bar('instance').createMenu('file', [
            {
              id: 'edit',
              label: 'edit',
              icon: 'edit',
              permissions: this.ctx.globals.getPermissions(['instancesUpdate']),
              click: () => {
                this.dialog.instance = true
              }
            },
            {
              id: 'delete',
              label: 'delete',
              icon: 'delete',
              permissions: this.ctx.globals.getPermissions(['instancesDelete']),
              click: () => {
                this.dialog.deleteInstance = true
              }
            }, 
          ]).createMenu('view', [
            {
              id: 'dashboard',
              label: 'Dashboard',
              permissions: this.ctx.globals.getPermissions(['instancesRead']),
              click: () => {
                  window.components.router().push({
                    path: '/process-manager/instances/dashboard',
                    query: {
                      service: this.$route.query.service,
                      instance: this.instanceUuid,
                    }
                  })
              },
            },
          ]).attrs({
            id: 'instance-toolbar',
          }),
          components.section().append(() => [
            components.grid().append(() => [
              components.cell(4).append(() => [
                components.list(`${this.ctx.i18n.i18n.t('instance')}: ${this.instanceSettings.name}`).attrs({ dense: false, flat: false }).append(() => [
                  components.item().append('section', () => [
                    processManager.getActionButtons(this.instanceStatus, (ident) => {
                      this.ctx.processManager.putInstanceAction(this.processManagerUuid, this.instanceSettings.uuid, ident)
                    }).attrs({ 
                      id: 'instance-card-action',
                      permissions: this.ctx.globals.getPermissions(['instancesUpdate']),
                    }),
                  ]),
                  components.item().append('section', () => [
                    processManager.getStatusComponents(this.instanceStatus),
                  ]).append('side', () => [
                    this.getInstanceActionMenu(this.instanceStatus).attrs({
                      permissions: this.ctx.globals.getPermissions(['instancesUpdate']),
                    }),
                  ]),
                  components.input('enabled', 'checkbox', this.instanceSettings.enabled, (val) => {
                    this.instanceSettings.enabled = val
                    this.ctx.processManager.putInstanceSettings(this.processManagerUuid, this.instanceUuid, this.instanceSettings)
                  }).attrs({
                    hint: 'enableAndAutoStart',
                    permissions: this.ctx.globals.getPermissions(['instancesUpdate']),
                  }),
                  components.item('processManager', this.processManager.name),
                  components.item('name', this.instanceStatus.name),
                  components.item('description', this.instanceStatus.description),
                  components.item('ID', this.instanceStatus.uuid),
                  components.item('portOffsetPrefix', this.ctx.i18n.i18n.t('displayPortOffsetPrefix', { portOffsetPrefix: this.instanceStatus.portOffsetPrefix })),
                  components.renderIf(this.instanceStatus.tsLastStart, components.item('tsLastStart', components.getDateString(this.instanceStatus.tsLastStart))),
                  components.renderIf(this.instanceStatus.tsLastStop, components.item('tsLastStop', components.getDateString(this.instanceStatus.tsLastStop))),
                  components.renderIf(this.instanceStatus.lastError, components.item('lastError', this.instanceStatus.lastError)),
                  components.renderIf(this.instanceSettings.tsLastError, components.item('tsLastError', components.getDateString(this.instanceStatus.tsLastError))),
                  components.renderIf(this.instanceServers.processes.length, components.item('servers', 'endpoints', true).attrs({
                    id: 'instance-card-servers',
                  }).on({ 
                    onClick: () => this.initServers()
                  }).append(() => [
                    components.list().append(() => [
                      ...this.instanceServers.processes.map((process) => {
                        return components.item(process.name, process.servers.map((server) => this.ctx.processManager.getEndpoint(server)).join(', ')).on({
                          clickable: process.enable,
                          permissions: this.ctx.globals.getPermissions(['processesRead']),
                          onClick: () => components.router().push({
                            path: '/process-manager/instances/processes/status/index',
                            query: {
                              service: this.$route.query.service,
                              instance: this.instanceUuid,
                              process: process.uuid,
                            }
                          })
                        })
                      })
                    ])
                  ])),
                  components.renderIf(!this.instanceServers.processes.length, components.item('servers', 'endpoints', true).append(() => [
                    components.list().append(() => [
                      components.item('noData'),
                    ])
                  ])),
                  components.expandItem('performance').append(() => [
                    components.chart('', this.chartOptions, this.timeSeries, 250),
                    components.button('', this.playTimeseries ? 'stop' : 'play_arrow').on({
                      onClick: () => this.handlePlayTimeseries()
                    }).attrs({
                      color: this.playTimeseries ? 'negative' : 'positive',
                      tooltip: 'chart',
                      flat: true,
                      dense: true,
                    }),
                  ]),
                ]),
              ]).attrs({
                id: 'instance-card',
              }),

              components.cell(8).append(() => [
                components.card('processes').attrs({ flat: !this.ctx.platform.browser }).append(() => [
                  components.table(
                    '', 
                    [
                      {
                        i18nLabel: '',
                        align: 'left',
                        name: 'list',
                        field: 'list',
                      },
                      {
                        i18nLabel: 'name',
                        align: 'left',
                        name: 'name',
                        field: 'name',
                        sortable: true,
                      },
                      {
                        i18nLabel: 'description',
                        align: 'left',
                        name: 'description',
                        field: 'description',
                        sortable: true,
                        style: 'max-width: 250px;',
                        classes: 'ellipsis',
                      },
                      {
                        i18nLabel: '',
                        align: 'left',
                        name: 'links',
                        field: 'links'
                      },
                      {
                        i18nLabel: 'status',
                        align: 'left',
                        name: 'status',
                        field: 'status',
                        sortable: true,
                      },
                      {
                        i18nLabel: 'action',
                        align: 'left',
                        name: 'action',
                        field: 'action'
                      },
                    ], 
                    this.instanceStatus.processes,
                  ).attrs({
                    id: 'processes-table',
                    selection: 'multiple',
                    selected: this.selectedProcesses,
                    rowKey: 'uuid',
                    grid: !this.ctx.platform.browser,
                    flat: true,
                    enableAddBtn: true,
                    enableDeleteBtn: true,
                    permissionsAdd: [29],
                    permissionsDelete: [32],
                    permissionsSelect: [32],
                    'onUpdate:selected': (val) => this.selectedProcesses = val, 
                    'onClick:row': (props) => components.router().push('/process-manager/instances/processes/index', { 
                      service: this.$route.query.service,
                      instance: this.instanceUuid,
                      process: props.row.uuid,
                    }),
                    loading: this.loading,
                    loadingLabel: 'fetchProcesses',
                    'onClick:add': () => {
                      this.dialog.addProcess = true
                    },
                    'onConfirm:delete': () => {
                      this.onDeleteSelected()
                    },
                    maxHeight: this.ctx.platform.browser ? '75vh' : undefined,
                  }).append('body-cell-list', (props) => [
                    this.getDropdown(props)
                  ]).append('body-cell-status', (props) => [
                    components.td().append(() => [
                      processManager.getStatusComponents(props.row, 'flex', true),
                    ])
                  ]).append('body-cell-name', (props) => [
                    components.td().append(() => [
                      components.flex().append(() => [
                        components.icon(
                          props.row.uuid === 'ProcessManager' ? 'home' : 'folder',
                          props.row.uuid === 'ProcessManager' ? 'primary' : 'grey',
                          'xs',
                        ),
                        components.text(props.value).attrs({ type: props.row.uuid === 'ProcessManager' ? 'bold' : '' })
                      ])
                    ])
                  ]).append('body-cell-action', (props) => [
                    components.td().append(() => [
                      this.getActionMenu(props, {
                        update: [31],
                        delete: [32],
                      }).attrs({
                        permissions: this.ctx.globals.getPermissions(['processesUpdate', 'processesDelete'])
                      })
                    ])
                  ]).append('body-cell-links', (props) => [
                    components.td().append(() => [
                      components.element('div').on({ onClick: this.withModifiers(() => {}, ['stop'])}).append(() => [
                       components.button('', 'web').on({
                          onClick: () => window.processManager.toModule(props.row.name, this.$route.query.service, this.instanceUuid, props.row.uuid)
                        }).attrs({
                         dense: true,
                         flat: true,
                         round: true,
                         disable: !window.processManager.hasModule(props.row.name),
                         tooltip: 'module',
                        }),
                        components.button('', 'settings').on({
                          onClick: () => window.processManager.toConfiguration(
                            this.$route.query.service, 
                            this.instanceUuid, 
                            props.row.uuid,
                            props.row.name,
                          )
                        }).attrs({
                         dense: true,
                         flat: true,
                         round: true,
                         tooltip: 'configuration',
                          permissions: this.ctx.globals.getPermissions(['processesUpdate']),
                        }),
                        components.button('', 'checklist_rtl').on({
                          onClick: () => window.processManager.toStatus(this.$route.query.service, this.instanceUuid, props.row.uuid)
                        }).attrs({
                         dense: true,
                         flat: true,
                         round: true,
                         tooltip: 'status',
                         permissions: this.ctx.globals.getPermissions(['processesUpdate']),
                        }),
                        components.button('', 'description').on({
                          onClick: () => window.processManager.toLogs(this.$route.query.service, this.instanceUuid, props.row.uuid)
                        }).attrs({
                         dense: true,
                         flat: true,
                         round: true,
                         tooltip: 'logs',
                         permissions: this.ctx.globals.getPermissions(['changelogRead']),
                        }),                       
                      ])

                    ])
                  ]).append('item', (props) => [
                    components.tdItem(components.card().append(() => [
                      components.renderIf(this.selectable, components.section().append(() => [
                        components.input('select', 'checkbox', props.selected, (val) => props.selected = val).attrs({
                          permissions: this.ctx.globals.getPermissions(['processesDelete']),
                        }),
                      ])),
                      components.section().append(() => [
                        components.item(props.row.name, props.row.name !== props.row.description && props.row.description ? props.row.description : '').attrs({
                          clickable: true,
                          onClick: () => components.router().push('/process-manager/instances/processes/index', { 
                            service: this.$route.query.service,
                            instance: this.instanceUuid,
                            process: props.row.uuid,
                          }),
                        }).append('side', () => [
                          this.getActionMenu(props),
                        ]),
                        components.item().append('section', () => [
                          processManager.getStatusBadge(props.row),
                        ]).append('side', () => [
                          this.getDropdown(props)
                        ]),
                      ]),
                    ]))
                  ])
                ]),
              ])
            ]),
          ]),

          components.dialogDelete(this.instanceSettings.name, this.dialog.deleteInstance, (val) => this.dialog.deleteInstance = val).on({
            onOk: () => {
              this.removeInstance()
            },
          }), 

          // Forms
          components.formDialog('instance', this.dialog.instance, (val) => this.dialog.instance = val, 'public/forms/instance.js').formOn({
            onSubmit: () => {
              this.dialog.instance = false
            },
            onCancel: () => this.dialog.instance = false,
          }).formAttrs({
            instanceUuid: this.instanceSettings.uuid,
          }).on({
            onHide: () => this.init(),
            onShow: () => this.stopPoll()
          }),

          components.formDialog('process', this.dialog.process, (val) => this.dialog.process = val, 'public/forms/process.js').formOn({
            onSubmit: () => {
              this.dialog.process = false
            },
            onCancel: () => {
              this.dialog.process = false
            },
          }).formAttrs({
            instanceUuid: this.instanceSettings.uuid,
            processUuid: this.processSettings.uuid,
          }).on({
            onHide: () => this.init(),
            onShow: () => this.stopPoll()
          }),
          
          // Add templates
          components.dialog('addProcess', this.dialog.addProcess, (val) => this.dialog.addProcess = val).append(() => [
            components.listVirtual('processes', 'clickToAdd', this.templates).attrs({
              filterCallback: (item, val) => item.name.toLowerCase().includes(val.toLowerCase()),
              ident: 'uuid',
            }).append('item', (item) => [
              components.item(item.name, item.description).attrs({ 
                clickable: !this.loading, 
                icon: 'add',
              }).on({
                onClick: () => {
                  this.loading = true
                  if (item.arguments) {
                    item.arguments.forEach((attr, idx) => {
                      if (attr.includes('-discovery')) {
						item.arguments[idx] = `-discovery ${this.apigatewayEndpoint}`
                      }
                    }) 
                  }
                  
                  this.ctx.processManager.putProcessSettings(this.processManagerUuid, this.instanceUuid, '', item).then(() => {
                    this.init()   
                  })
                }
              })
            ]),     
          ])
        ]),
      ]
    },
  }
})