// Copyright 2021 MST Systemtechnik AG. All Rights Reserved.

window.addEventListener('load', function () {

    'use strict';

    // Error panel
    let modalTimer = 0;
    function showError(title, text, displayDuration) {
        let infoPanelEl = document.getElementById('error-panel');
        infoPanelEl.innerHTML = '<b>' + title + '</b><br>';
        infoPanelEl.innerHTML += text;
        infoPanelEl.className = 'sticky';
        // Start timer for auto close
        if (modalTimer) {
            clearTimeout(modalTimer);
        }
        if (displayDuration) {
            modalTimer = setTimeout(function () {
                infoPanelEl.className = 'sticky invisible';
            }, displayDuration);
        }
    }
    
    // Status request handling
    let infoRequest = null;
    let infoTimer = 0;

    function getXmlHttpinfoRequestObject() {
        if (!infoRequest) {
            // Create a new XMLHttpRequest object
            infoRequest = new XMLHttpRequest();
        }
        return infoRequest;
    }

    function pollTopInformations() {
        if (infoTimer) {
            clearTimeout(infoTimer);
            infoTimer = 0;
        }
        let now = new Date();
        infoRequest = getXmlHttpinfoRequestObject();
        let url = '/v1/top/status?_' + now.getTime();
        infoRequest.onreadystatechange = function () {
            // Check response is ready or not
            if (infoRequest.readyState === 4) {
                if (infoRequest.status === 200) {
                    // Create / update status list
                    updateInformations(JSON.parse(infoRequest.responseText));
                    // Update the live data every second 
                    infoTimer = setTimeout(pollTopInformations, 1000);
                } else {
                    // error-status view
                    let statusText = infoRequest.status + ' - ' + infoRequest.statusText;
                    if (infoRequest.status === 0) {
                        statusText = 'No Connection';
                    }
                    showError('Status Poll', 'Error ' + statusText + '<br><br>' + infoRequest.responseText, 6000);
                    // Retry in 5 sec
                    infoTimer = setTimeout(pollTopInformations, 5000);
                }
            }
        };
        // asynchronous requests
        infoRequest.open("GET", url, true);
        infoRequest.timeout = 5000;
        // Send the request over the network
        infoRequest.send();
    }

    // Action functions
    function doAction(pid, action) {
        let now = new Date();
        let url = '/v1/top/actions/' + pid;
        url += '?action=' + action;
        url += '&_' + now.getTime();
        let actionRequest = new XMLHttpRequest();
        actionRequest.onreadystatechange = function () {
            if (actionRequest.readyState === 2) {
                // showError('Action ' + action, 'Sent', 2000);
            } else if (actionRequest.readyState === 4) {
                if (actionRequest.status === 200) {
                    // showError('Action ' + action, 'Ok', 1000);
                } else {
                    showError('Action ' + action, 'failed, code ' + actionRequest.status + ' - ' + actionRequest.statusText + '<br><br>' + actionRequest.responseText, 5000);
                }
            }
        };
        actionRequest.open("PUT", url, true);
        actionRequest.timeout = 5000;
        actionRequest.send();
    }

    // Treeview switch
    let treeView = false;
    document.getElementById('switch-treeview').addEventListener('change', function () {
        treeView = !treeView;
        pollTopInformations();
    });

    // Add process table header
    let sortedHeader = 'processName';
    let sortedAsc = true;
    function addProcessTableHeader() {
        let processTbl = document.createElement('table');
        let processTblHdr = document.createElement('thead');
        let processTableHeaderDef = [['','Kill'], ['pid','PID'], ['userName','User'], ['cpuUsagePercent','Pri'], ['cpuUsagePercent','CPU %'], 
                    ['usedMemory','Mem MB'], ['cpuTimeSec','CpuTime'], ['upTimeSec','UpTime'], ['threadCount','Treads'], ['handleCount','Handles'], 
                    ['diskUsage','Disk MB/s'], ['processName','Process'], ['parentPid','ParentPid'], ['commandLine','CmdLine']];

        processTableHeaderDef.forEach(function (headerDef) {
            let thEl = document.createElement('th');
            let objField = headerDef[0];
            if (objField !== '')
            {
                thEl.className = 'th-sortable';
                if (objField === 'processName') {
                    thEl.className = 'th-sortable th-sorted';
                }
                thEl.addEventListener('click', function () {
                    if (sortedHeader === objField) {
                        sortedAsc = !sortedAsc;
                    }
                    else {
                        document.querySelector('.th-sorted').className = 'th-sortable';
                        thEl.className = 'th-sortable th-sorted';
                        sortedHeader = headerDef[0];
                        sortedAsc = true;
                    }
                    pollTopInformations();
                });
            }
            processTblHdr.appendChild(thEl).appendChild(document.createTextNode(headerDef[1]));
        });
        processTbl.appendChild(processTblHdr);
        // Append to formular
        let processesEl = document.getElementById('processes');
        processesEl.appendChild(processTbl);
    }
    addProcessTableHeader();

    // Start first poll
    pollTopInformations();

    function addActionButton(pid, ptext, text, action) {
        let btn = document.createElement('button');
        btn.innerHTML = text;
        btn.className = 'btn-action';
        btn.addEventListener('click', function () {
            if (confirm('Are you sure to kill process "' + ptext + '" (pid ' + pid + ') ?')) {
                doAction(pid, action);
            }
        });
        return btn;
    }

    function secToString(sec) {
        return [3600*24, 3600, 60]
            .reduceRight(
                (p, b) => r => [Math.floor(r / b)].concat(p(r % b)),r => [r]
            )(sec)
            .map(a => a.toString().padStart(2, '0'))
            .join(':')
            .replace(/^00:00:00/, '--:--:--').replace(/^00:00/, '--:--').replace(/^00/, '--');
    }

    const sortBy = (field, reverse) => {
        reverse = !reverse ? 1 : -1;
        return function(a, b) {
            a = a[field];
            if (typeof a === "string") {
                a = a.toLowerCase();
            }
            b = b[field];
            if (typeof b === "string") {
                b = b.toLowerCase();
            }
            return reverse * ((a > b) - (b > a));
        }
    }

    function insertCell(tblRowEl, fieldName, text) {
        let cellEl = tblRowEl.insertCell();
        cellEl.className = 'process-cell-' + fieldName;
        cellEl.innerHTML = text;
        return cellEl;
    }

    // Helper to build treeview
    function addChildWithProcesses(processObj, prefix, processes, treeviewProcesses) {
        let suffix = '';
        if (prefix !== '') {
            suffix = '&rdca;&nbsp;';
        }
        processObj.processName = prefix + suffix + processObj.processName;
        treeviewProcesses.push(processObj);
        if (prefix === '') {
            prefix = ' ';
        }
        else {
            prefix += '&nbsp;&nbsp;';
        }
        let pIdx = -1
        while ((pIdx = processes.findIndex(x => x.parentPid === processObj.pid)) != -1) {
            let childProcess = processes[pIdx];
            processes.splice(pIdx, 1);
            addChildWithProcesses(childProcess, prefix, processes, treeviewProcesses);
        }
    }

    let markedPid = -1;
    function updateInformations(informations) {

        let mainInfoEl = document.getElementById('main-info-txt');
        let mainInfoStr = 'CpuName:<b>' + informations.mainInfo.cpuName + 
            '</b>, ProductName:<b>' + informations.mainInfo.productName +
            '</b>, Architecture:<b>' + informations.mainInfo.architecture +
            '</b>, ComputerName:<b>' + informations.mainInfo.computerName + 
            '</b>, Uptime:<b>' + secToString(informations.mainInfo.uptimeSec) + 
            '</b>, ProcessCount:<b>' + informations.mainInfo.processCount +
            '</b>, Running:<b>' + informations.mainInfo.processRunning + '</b>';
        mainInfoEl.innerHTML = mainInfoStr;

        let cpuUsageBarEl = document.getElementById('cpu-usage-bar');
        cpuUsageBarEl.style.width = informations.mainInfo.cpuUsagePercent + '%';
        cpuUsageBarEl.innerHTML = Number.parseFloat(informations.mainInfo.cpuUsagePercent).toFixed(0) + '%';
        if (informations.mainInfo.cpuUsagePercent > 90) {
            cpuUsageBarEl.className = 'progress-bar progress-bar-high-high';
        }
        else if (informations.mainInfo.cpuUsagePercent > 80) {
            cpuUsageBarEl.className = 'progress-bar progress-bar-high';
        }
        else  {
            cpuUsageBarEl.className = 'progress-bar';
        }

        let memUsageBarEl = document.getElementById('mem-usage-bar');
        memUsageBarEl.style.width = informations.mainInfo.usedMemoryPercent + '%';
        memUsageBarEl.innerHTML = Number.parseFloat(informations.mainInfo.usedMemoryPercent).toFixed(0) + '%';
        if (informations.mainInfo.usedMemoryPercent > 90) {
            memUsageBarEl.className = 'progress-bar progress-bar-high-high';
        }
        else if (informations.mainInfo.usedMemoryPercent > 80) {
            memUsageBarEl.className = 'progress-bar progress-bar-high';
        }
        else  {
            memUsageBarEl.className = 'progress-bar';
        }
        let memUsageEl = document.getElementById('mem-usage');
        memUsageEl.innerHTML = Number.parseFloat(informations.mainInfo.usedMemory / 1024 / 1024 / 1024).toFixed(2) + 'G / ' + 
            Number.parseFloat(informations.mainInfo.totalMemory / 1024 / 1024 / 1024).toFixed(2) + 'G';

        let swapUsageBarEl = document.getElementById('swap-usage-bar');
        swapUsageBarEl.style.width = informations.mainInfo.usedSwapPercent + '%';
        swapUsageBarEl.innerHTML = Number.parseFloat(informations.mainInfo.usedSwapPercent).toFixed(0) + '%';
        if (informations.mainInfo.usedSwapPercent > 80) {
            swapUsageBarEl.className = 'progress-bar progress-bar-high-high';
        }
        else if (informations.mainInfo.usedSwapPercent > 70) {
            swapUsageBarEl.className = 'progress-bar progress-bar-high';
        }
        else  {
            swapUsageBarEl.className = 'progress-bar';
        }
        let swapUsageEl = document.getElementById('swap-usage');
        swapUsageEl.innerHTML = Number.parseFloat(informations.mainInfo.usedSwap / 1024 / 1024 / 1024).toFixed(2) + 'G / ' + 
            Number.parseFloat(informations.mainInfo.totalSwap / 1024 / 1024 / 1024).toFixed(2) + 'G';
    
        // Sort process data
        informations.processes.sort(sortBy(sortedHeader, !sortedAsc));

        // Treeview ?
        if (treeView) {
            let treeviewProcesses = [];

            while (informations.processes.length > 0) {
                // Find root for this process
                let rootObj = informations.processes[0];
                let pIdx = -1;
                while ((pIdx = informations.processes.findIndex(x => x.pid === rootObj.parentPid)) != -1) {
                    rootObj = informations.processes[pIdx];
                }
                // Add root and childs
                let prefix = '';
                addChildWithProcesses(rootObj, prefix, informations.processes, treeviewProcesses);
                // Remove this root from list
                pIdx = informations.processes.findIndex(x => x.pid === rootObj.pid);
                if (pIdx != -1) {
                    informations.processes.splice(pIdx, 1);
                }
            }
            
            informations.processes = treeviewProcesses;
        }

        // Clear and rebuild table due to
        // - new sorted 
        // - process no more present
        let tbodyEl = document.querySelector('tbody');
        if (tbodyEl !== null) {
            tbodyEl.innerHTML = '';
        }
        // Iterate all processes
        informations.processes.forEach(function (processObj) {
            let processTblRow = document.querySelector('table').insertRow();
            processTblRow.id = 'process-row-' + processObj.pid;
            processTblRow.addEventListener('click', function () {
                markedPid = processObj.pid;
                processTblRow.className = 'row-marked';
                let oldMarkedEl = document.querySelector('.row-marked');
                if (oldMarkedEl !== null) {
                    oldMarkedEl.className = '';
                }
            });

            if (processObj.pid === markedPid) {
                processTblRow.className = 'row-marked';
            }

            // Add kill button
            processTblRow.insertCell().appendChild(addActionButton(processObj.pid, processObj.processName, 'X', 'kill'));

            // Add other cells
            insertCell(processTblRow, 'pid', processObj.pid, processObj.pid);
            insertCell(processTblRow, 'userName', processObj.userName);
            insertCell(processTblRow, 'priority', processObj.priority);
            let cpuUsageEl = insertCell(processTblRow, 'cpuUsagePercent', Number.parseFloat(processObj.cpuUsagePercent).toFixed(1));
            if (processObj.cpuUsagePercent < 0.1) {
                cpuUsageEl.classList.add('usage-zero');
            }
            insertCell(processTblRow, 'usedMemory', Number.parseFloat(processObj.usedMemory / 1024 / 1024).toFixed(2));
            insertCell(processTblRow, 'cpuTimeSec', secToString(processObj.cpuTimeSec));
            insertCell(processTblRow, 'upTimeSec',  secToString(processObj.upTimeSec));
            insertCell(processTblRow, 'threadCount', processObj.threadCount);
            insertCell(processTblRow, 'handleCount', processObj.handleCount);
            let diskUsage = processObj.diskUsage / 1024 / 1024;
            let diskUsageEl = insertCell(processTblRow, 'diskUsage', Number.parseFloat(diskUsage).toFixed(2));
            if (diskUsage < 0.01) {
                diskUsageEl.classList.add('usage-zero');
            }
            insertCell(processTblRow, 'processName', processObj.processName);
            insertCell(processTblRow, 'parentPid', processObj.parentPid);
            if (processObj.parentPidUnknown) {
                processTblRow.querySelector('.process-cell-parentPid').classList.add('unknown_parent_pid');
            }
            insertCell(processTblRow, 'commandLine', processObj.commandLine);
        });
    }

});